import logging

from prometheus_client.twisted import MetricsResource
from prometheus_client import Counter, Summary, Gauge
from twisted.web.server import Site
from twisted.web import server, resource
from twisted.internet import task
from scrapy.exceptions import NotConfigured
from scrapy.utils.reactor import listen_tcp
from scrapy import signals

logger = logging.getLogger(__name__)


class WebService(Site):
    """

    """
    def __init__(self, crawler):
        if not crawler.settings.getbool('PROMETHEUS_ENABLED', True):
            raise NotConfigured
        self.tasks = []
        self.stats = crawler.stats
        self.crawler = crawler
        self.name = crawler.settings.get('BOT_NAME')
        self.port = crawler.settings.get('PROMETHEUS_PORT', [9410])
        self.host = crawler.settings.get('PROMETHEUS_HOST', '0.0.0.0')
        self.path = crawler.settings.get('PROMETHEUS_PATH', 'metrics')
        self.interval = crawler.settings.get('PROMETHEUS_UPDATE_INTERVAL', 30)

        self.spr_item_scraped = Gauge(
            'spr_items_scraped', 'Spider items scraped', ['spider'])
        self.spr_item_dropped = Gauge(
            'spr_items_dropped', 'Spider items dropped', ['spider'])
        self.spr_response_received = Gauge(
            'spr_response_received', 'Spider responses received', ['spider'])
        self.spr_opened = Gauge('spr_opened', 'Spider opened', ['spider'])
        self.spr_closed = Gauge(
            'spr_closed', 'Spider closed', ['spider', 'reason'])

        self.spr_downloader_request_bytes = Gauge(
            'spr_downloader_request_bytes', '...', ['spider'])
        self.spr_downloader_request_total = Gauge(
            'spr_downloader_request_total', '...', ['spider'])
        self.spr_downloader_request_count = Gauge(
            'spr_downloader_request', '...', ['spider', 'method'])
        self.spr_downloader_response_count = Gauge(
            'spr_downloader_response', '...', ['spider'])
        self.spr_downloader_response_bytes = Gauge(
            'spr_downloader_response_bytes', '...', ['spider'])
        self.spr_downloader_response_status_count = Gauge(
            'spr_downloader_response_status', '...', ['spider', 'code'])

        self.spr_log_count = Gauge('spr_log', '...', ['spider', 'level'])

        self.spr_duplicate_filtered = Gauge(
            'spr_duplicate_filtered', '...', ['spider'])

        self.spr_memdebug_gc_garbage_count = Gauge(
            'spr_memdebug_gc_garbage', '...', ['spider'])
        self.spr_memdebug_live_refs = Gauge(
            'spr_memdebug_live_refs', '...', ['spider'])
        self.spr_memusage_max = Gauge(
            'spr_memusage_max', '...', ['spider'])
        self.spr_memusage_startup = Gauge(
            'spr_memusage_startup', '...', ['spider'])

        self.spr_scheduler_dequeued = Gauge(
            'spr_scheduler_dequeued', '...', ['spider'])
        self.spr_scheduler_enqueued = Gauge(
            'spr_scheduler_enqueued', '...', ['spider'])
        self.spr_scheduler_enqueued_memory = Gauge(
            'spr_scheduler_enqueued_memory', '...', ['spider'])

        self.spr_offsite_domains_count = Gauge(
            'spr_offsite_domains', '...', ['spider'])
        self.spr_offsite_filtered_count = Gauge(
            'spr_offsite_filtered', '...', ['spider'])

        self.spr_request_depth = Gauge(
            'spr_request_depth', '...', ['spider'])
        self.spr_request_depth_max = Gauge(
            'spr_request_depth_max', '...', ['spider'])

        root = resource.Resource()
        self.promtheus = None
        root.putChild(self.path.encode('utf-8'), MetricsResource())
        server.Site.__init__(self, root)

        crawler.signals.connect(self.engine_started, signals.engine_started)
        crawler.signals.connect(self.engine_stopped, signals.engine_stopped)

        crawler.signals.connect(self.spider_opened, signals.spider_opened)
        crawler.signals.connect(self.spider_closed, signals.spider_closed)
        crawler.signals.connect(self.item_scraped, signals.item_scraped)
        crawler.signals.connect(self.item_dropped, signals.item_dropped)
        crawler.signals.connect(self.response_received,
                                signals.response_received)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    def engine_started(self):
        # Start server endpoint for exporting metrics
        self.promtheus = listen_tcp(self.port, self.host, self)

        # Periodically update the metrics
        tsk = task.LoopingCall(self.update)
        self.tasks.append(tsk)
        tsk.start(self.interval, now=True)

    def engine_stopped(self):
        # Stop all periodic tasks
        for tsk in self.tasks:
            if tsk.running:
                tsk.stop()

        # Stop metrics exporting
        self.promtheus.stopListening()

    def spider_opened(self, spider):
        self.spr_opened.labels(spider=self.name).inc()

    def spider_closed(self, spider, reason):
        self.spr_closed.labels(spider=self.name, reason=reason).inc()

    def item_scraped(self, item, spider):
        self.spr_item_scraped.labels(spider=self.name).inc()

    def response_received(self, spider):
        self.spr_response_received.labels(spider=self.name).inc()

    def item_dropped(self, item, spider, exception):
        self.spr_item_scraped.labels(spider=self.name).inc()

    def update(self):
        logging.debug(self.stats.get_stats())

        # Downloader Request Stats
        self.request_stats()

        # Downloader Request Stats
        self.response_stats()

        # Logging Stats
        self.logging_stats()

        # Memory Debug Stats
        self.memory_debug_stats()

        # Memory Usage Stats
        self.memory_usage_stats()

        # Scheduler Stats
        self.scheduler_stats()

        # Off-Site Filtering Stats
        self.offsite_stats()

        # Duplicate Stats
        self.duplicate_filter_stats()

        # Request Depth
        self.request_depth()

    def request_depth(self):
        depth = self.stats.get_value('request_depth_max', 0)
        self.spr_request_depth_max.labels(spider=self.name).set(depth)
        for i in range(depth):
            stat = 'request_depth_count/{}'.format(i)
            depthv = self.stats.get_value(stat, 0)
            self.spr_request_depth.labels(spider=self.name).set(depthv)

    def duplicate_filter_stats(self):
        dup = self.stats.get_value('dupefilter/filtered', 0)
        self.spr_duplicate_filtered.labels(spider=self.name).set(dup)

    def memory_debug_stats(self):
        mdgc_count = self.stats.get_value('memdebug/gc_garbage_count', 0)
        self.spr_memdebug_gc_garbage_count.labels(
            spider=self.name).set(mdgc_count)

        mdlr_count = self.stats.get_value('memdebug/live_refs/MySpider', 0)
        self.spr_memdebug_live_refs.labels(
            spider=self.name).set(mdlr_count)

    def memory_usage_stats(self):
        mum_count = self.stats.get_value('memusage/max', 0)
        self.spr_memusage_max.labels(spider=self.name).set(mum_count)

        mus_count = self.stats.get_value('memusage/startup', 0)
        self.spr_memusage_startup.labels(spider=self.name).set(mus_count)

    def scheduler_stats(self):
        dequeued = self.stats.get_value('scheduler/dequeued', 0)
        self.spr_scheduler_dequeued.labels(spider=self.name).set(dequeued)

        enqueued = self.stats.get_value('scheduler/enqueued', 0)
        self.spr_scheduler_enqueued.labels(spider=self.name).set(enqueued)

        enqueued_mem = self.stats.get_value('scheduler/enqueued/memory', 0)
        self.spr_scheduler_enqueued_memory.labels(
            spider=self.name).set(enqueued_mem)

        dequeued_mem = self.stats.get_value('scheduler/dequeued/memory', 0)
        self.spr_scheduler_enqueued_memory.labels(
            spider=self.name).set(dequeued_mem)

    def offsite_stats(self):
        od_count = self.stats.get_value('offsite/domains', 0)
        self.spr_offsite_domains_count.labels(
            spider=self.name).set(od_count)

        of_count = self.stats.get_value('offsite/filtered', 0)
        self.spr_offsite_filtered_count.labels(
            spider=self.name).set(of_count)
        
    def request_stats(self):
        for i in ['GET', 'PUT', 'DELETE', 'POST']:
            stat = 'downloader/request_method_count/{}'.format(i)
            count = self.stats.get_value(stat, 0)
            if count > 0:
                self.spr_downloader_request_count.labels(
                    spider=self.name, method=i).set(count)

        total_count = self.stats.get_value('downloader/request_count', 0)
        self.spr_downloader_request_total.labels(
            spider=self.name).set(total_count)

        request_bytes = self.stats.get_value('downloader/request_bytes', 0)
        self.spr_downloader_request_bytes.labels(
            spider=self.name).set(request_bytes)

    def response_stats(self):
        response_count = self.stats.get_value('downloader/response_count', 0)
        self.spr_downloader_response_count.labels(
            spider=self.name).set(response_count)

        for i in ['200', '404', '500']:
            stat = 'downloader/response_status_count/{}'.format(i)
            status = self.stats.get_value(stat, 0)
            self.spr_downloader_response_status_count.labels(
                spider=self.name, code=i).set(status)

        response_bytes = self.stats.get_value('downloader/response_bytes', 0)
        self.spr_downloader_response_bytes.labels(
            spider=self.name).set(response_bytes)

    def logging_stats(self):
        for i in ['DEBUG', 'ERROR', 'INFO', 'CRITICAL', 'WARNING']:
            level = self.stats.get_value('log_count/{}'.format(i), 0)
            self.spr_log_count.labels(
                spider=self.name, level=i).set(level)



import csv
import logging

from opennem.core.loader import load_data
from opennem.core.normalizers import clean_float, normalize_duid
from opennem.db import SessionLocal
from opennem.db.models.opennem import Facility
from opennem.importer.mms import mms_import

logger = logging.getLogger(__name__)


def import_mms_emissions() -> None:
    # mms = mms_import()

    facility_poll_map = {
        "APS": 1.228625,
        "CALL_A_4": 0.93134349,
        "CALL_B_1": 0.93134349,
        "CALL_B_2": 0.93134349,
        "SWAN_E": 0.42795234,
        "OSB-AG": 0.544686,
        "HWPS1": 1.558001,
        "HWPS2": 1.558001,
        "HWPS3": 1.558001,
        "HWPS4": 1.558001,
        "COLNSV_3": 1.191769,
        "MM3": 1.162987,
        "MM4": 1.162987,
        "NPS1": 1.130635,
        "NPS2": 1.130635,
        "NPSNL2": 1.130635,
        "PLAYFB1": 1.509041,
        "PLAYFB2": 1.509041,
        "PLAYFB3": 1.509041,
        "PLAYFB4": 1.509041,
        "PLAYB-AG": 1.509041,
        "COLNSV_4": 1.191769,
        "COLNSV_5": 1.191769,
        "HWPS7": 1.558001,
        "HWPS8": 1.558001,
        "SWAN_B_1": 1.091,
        "SWAN_B_2": 1.091,
        "SWAN_B_3": 1.091,
        "SWAN_B_4": 1.091,
        "CALL_A_2": 1.328,
        "CALL_A_4": 1.328,
        "COLNSV_1": 1.191769,
        "COLNSV_2": 1.191769,
        "HWPS5": 1.558001,
        "HWPS6": 1.558001,
        "WW7": 1.030046,
        "WW8": 1.030046,
        # from wem
        "KWINANA_C5": 0.877,
        "KWINANA_C5": 0.877,
        # from reports
        "REDBANK1": 1.2,
        # from https://web.archive.org/web/20111002115520/http://carma.org/company/detail/6264
        "MOR1": 2.79,
        "MOR2": 2.79,
        "MOR3": 2.79,
    }

    # for station in mms:
    #     for facility in station.facilities:
    #         if facility.emissions_factor_co2 and not facility.code.endswith(
    #             "NL1"
    #         ):
    #             facility_poll_map[facility.code] = {
    #                 "intensity": facility.emissions_factor_co2,
    #                 "name": station.name,
    #                 "fueltech": facility.fueltech.code
    #                 if facility.fueltech
    #                 else "",
    #             }

    session = SessionLocal()

    for fac_code, pol_value in facility_poll_map.items():
        facility = (
            session.query(Facility)
            .filter_by(code=fac_code)
            .filter_by(network_id="NEM")
            .one_or_none()
        )

        if not facility:
            logger.info(
                "could not find facility: {} {}".format(fac_code, pol_value)
            )

            continue

        facility.emissions_factor_co2 = clean_float(pol_value)
        session.add(facility)
        session.commit()

    return None


def import_emissions_csv() -> None:
    EMISSION_MAPS = [
        {"filename": "wem_emissions.csv", "network": "WEM"},
        {"filename": "nem_emissions.csv", "network": "NEM"},
    ]

    for m in EMISSION_MAPS:
        import_emissions_map(m["filename"], m["network"])


def import_emissions_map(file_name: str, network_id: str) -> None:
    session = SessionLocal()

    content = load_data(file_name, from_project=True, skip_loaders=True)

    csv_content = content.splitlines()
    csvreader = csv.DictReader(csv_content)

    for rec in csvreader:
        facility_code = normalize_duid(rec["facility_code"])
        emissions_intensity = clean_float(rec["emissions_factor_co2"])

        if not facility_code:
            logger.info("No emissions intensity for {}".format(facility_code))
            continue

        facility = (
            session.query(Facility)
            .filter_by(code=facility_code)
            .filter_by(network_id=network_id)
            .one_or_none()
        )

        if not facility:
            logger.info("No stored facility for {}".format(facility_code))
            continue

        facility.emissions_factor_co2 = emissions_intensity
        session.add(facility)
        logger.info(
            "Updated {} to {}".format(facility_code, emissions_intensity)
        )

    session.commit()

    return None


if __name__ == "__main__":
    import_emissions_csv()
    import_mms_emissions()



import io
from typing import Any, Dict, Generator, Optional

import scrapy
from scrapy.http import Response

from opennem.pipelines.nem import (
    ExtractCSV,
    ReadStringHandle,
    TableRecordSplitter,
    UnzipSingleFilePipeline,
)
from opennem.utils.dates import get_date_component


class NemXLSSpider(scrapy.Spider):

    start_url: Optional[str] = None

    url_params = {
        "day": get_date_component("%d"),
        "month": get_date_component("%m"),
        "year": get_date_component("%Y"),
    }

    def start_requests(self) -> Generator[scrapy.Request, None, None]:
        if not self.start_url:
            return None

        request_url = self.start_url.format(**self.url_params)

        yield scrapy.Request(request_url)

    def parse(
        self, response: Response
    ) -> Generator[Dict[str, Any], None, None]:
        yield {"content": response.text}


class NemSingleMMSSpider(scrapy.Spider):

    pipelines = set(
        [
            UnzipSingleFilePipeline,
            ReadStringHandle,
            ExtractCSV,
            TableRecordSplitter,
        ]
    )

    url: Optional[str] = None

    def start_requests(self) -> Generator[scrapy.Spider, None, None]:
        if not hasattr(self, "url") or not self.url:
            raise Exception("{} requires url parameter".format(self.__class__))

        yield scrapy.Request(self.url)

    def parse(
        self, response: Response
    ) -> Generator[Dict[str, Any], None, None]:
        yield {"body_stream": io.BytesIO(response.body)}



# pylint: disable=no-member
"""
mv_facilities_all update mv_facilities_all with new balancing
summary mv

Revision ID: 98e2968b63e4
Revises: d0089ded02af
Create Date: 2021-01-14 06:27:10.956571

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "98e2968b63e4"
down_revision = "d0089ded02af"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
        drop materialized view if exists mv_facility_all;
        create materialized view mv_facility_all as
        select
            time_bucket_gapfill('1 hour', fs.trading_interval) as trading_interval,
            f.code,
            f.fueltech_id,
            f.network_id,
            f.network_region,
            f.interconnector,
            f.interconnector_region_to,
            max(fs.energy) as energy,
            case when avg(bs.price) >= 0  and min(fs.energy) >= 0 then
                coalesce(
                    max(fs.energy) * avg(bs.price),
                    0.0
                )
            else 0.0
            end as market_value,
            case when avg(f.emissions_factor_co2) >= 0  and min(fs.energy) >= 0 then
                coalesce(
                    max(fs.energy) * avg(f.emissions_factor_co2),
                    0.0
                )
            else 0.0
            end as emissions
        from mv_facility_energy_hour fs
            left join facility f on fs.facility_code = f.code
            left join mv_balancing_summary_region_hour bs on
                fs.trading_interval = bs.trading_interval and
                fs.network_id = bs.network_id and
                bs.network_region = f.network_region
        where
            f.fueltech_id is not null and
            fs.trading_interval <= now() and
            fs.trading_interval > '1900-01-01'
        group by
            1,
            f.code,
            f.fueltech_id,
            f.network_id,
            f.network_region,
            f.interconnector,
            f.interconnector_region_to
        order by 1 desc;
    """
    )


def downgrade() -> None:
    pass



import logging
from typing import List, Optional

from opennem.core.loader import load_data
from opennem.schema.time import TimeInterval, TimePeriod

logger = logging.getLogger(__name__)


def load_intervals() -> List[TimeInterval]:
    interval_dicts = load_data("intervals.json")

    intervals = [TimeInterval(**i) for i in interval_dicts]

    return intervals


def load_periods() -> List[TimePeriod]:
    period_dicts = load_data("periods.json")

    periods = [TimePeriod(**i) for i in period_dicts]

    return periods


INTERVALS = load_intervals()

INTERVALS_SUPPORTED = [i.interval_human for i in INTERVALS]

PERIODS = load_periods()

PERIODS_SUPPORTED = [i.period_human for i in PERIODS]


def get_interval_by_size(interval_size: int) -> Optional[TimeInterval]:
    """
    Get an interval by size

    """
    interval_lookup = list(
        filter(lambda x: x.interval == interval_size, INTERVALS)
    )

    if interval_lookup:
        return interval_lookup.pop()

    logger.error("Invalid interval {} not mapped".format(interval_size))

    return None


def get_interval(interval_human: str) -> TimeInterval:
    interval_lookup = list(
        filter(lambda x: x.interval_human == interval_human, INTERVALS)
    )

    if interval_lookup:
        return interval_lookup.pop()

    raise Exception("Invalid interval {} not mapped".format(interval_human))


def get_period(period_human: str) -> TimePeriod:
    period_lookup = list(
        filter(lambda x: x.period_human == period_human, PERIODS)
    )

    if period_lookup:
        return period_lookup.pop()

    raise Exception("Invalid interval {} not mapped".format(period_human))



"""
    mime module - get mime types from url (filename) or bytes


"""

import logging
import mimetypes
from io import BytesIO
from typing import Optional

try:
    import magic

    HAVE_MAGIC = True
except ImportError:
    HAVE_MAGIC = False

mimetypes.init()

logger = logging.getLogger(__name__)

CONTENT_TYPES = ["utf-8", "utf-8-sig", "latin-1"]


def mime_from_content(content: BytesIO) -> Optional[str]:
    """
    Use libmime to get mime type from content stream
    """
    if not HAVE_MAGIC:
        return None

    if isinstance(content, bytes):
        content = BytesIO(content)

    try:
        file_mime = magic.from_buffer(content.read(2048), mime=True)
    except Exception as e:
        logger.error("Error parsing mime from content: {}".format(e))
        return None

    logger.debug("Using magic to get mime type {}".format(file_mime))

    return file_mime


def mime_from_url(url: str) -> Optional[str]:
    """
    Use python mimetypes dir to get mime type from file extension
    """
    file_mime, _ = mimetypes.guess_type(url)
    logger.debug("Using mimetypes to get mime type {}".format(file_mime))
    return file_mime


def decode_bytes(content: bytes) -> str:
    """
    Brute force decode content
    """

    for enc in CONTENT_TYPES:
        try:
            return content.decode(enc)
        except Exception:
            pass

    raise Exception("Could not decode")



from opennem.db import get_database_engine
from opennem.utils.dates import subtract_days

TIMESCALE_VIEWS = [
    "mv_facility_energy_hour",
    "mv_nem_facility_power_5min",
    "mv_interchange_energy_nem_region",
    "mv_wem_facility_power_30min",
]

MATERIAL_VIEWS = ["mv_facility_all", "mv_region_emissions"]


def refresh_timescale_views(all: bool = False) -> None:
    """refresh timescale views"""
    __query = """
    CALL refresh_continuous_aggregate(
        continuous_aggregate => '{view}',
        window_start         => {date_from},
        window_end           => NULL
    );
    """

    engine = get_database_engine()

    dt = subtract_days(days=7)
    date_from = "'{}'".format(dt.strftime("%Y-%m-%d"))

    if all:
        date_from = "NULL"

    with engine.connect() as c:
        for v in TIMESCALE_VIEWS:
            query = __query.format(view=v, date_from=date_from)
            c.execution_options(isolation_level="AUTOCOMMIT").execute(query)


def refresh_material_views() -> None:
    """Refresh material views"""
    __query = "REFRESH MATERIALIZED VIEW CONCURRENTLY {view} with data"

    engine = get_database_engine()

    with engine.connect() as c:
        for v in MATERIAL_VIEWS:
            query = __query.format(view=v)
            c.execution_options(isolation_level="AUTOCOMMIT").execute(query)


def refresh_views() -> None:
    refresh_timescale_views(True)
    refresh_material_views()


if __name__ == "__main__":
    refresh_views()



# from opennem.pipelines.bulk_insert import BulkInsertPipeline
# from opennem.pipelines.csv import RecordsToCSVPipeline
from opennem.pipelines.nem.opennem import NemwebUnitScadaOpenNEMStorePipeline
from opennem.spiders.nemweb import NemwebSpider


class NemwebLatestTradingIS(NemwebSpider):
    name = "au.nem.latest.trading_is"
    start_url = "http://nemweb.com.au/Reports/Current/TradingIS_Reports/"
    limit = 2

    # only get most recent
    # limit = 1

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            # BulkInsertPipeline,
            # RecordsToCSVPipeline,
        ]
    )


class NemwebCurrentTradingIS(NemwebSpider):
    name = "au.nem.current.trading_is"
    start_url = "http://nemweb.com.au/Reports/Current/TradingIS_Reports/"
    limit = 0

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            # BulkInsertPipeline,
            # RecordsToCSVPipeline,
        ]
    )


class NemwebArchiveTradingIS(NemwebSpider):
    name = "au.nem.archive.trading_is"
    start_url = "http://nemweb.com.au/Reports/Archive/TradingIS_Reports/"
    limit = 0

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            # BulkInsertPipeline,
            # RecordsToCSVPipeline,
        ]
    )

    # Archives tend to contain large zips of embedded zips so throttle
    # to limit memory use
    custom_settings = {
        "CONCURRENT_REQUESTS": 1,
        "CONCURRENT_ITEMS": 1,
    }



# pylint: disable=no-member
"""
create vw_region_flow_emissions

Revision ID: 47b8cabac4bb
Revises: 6100cbbcaf7e
Create Date: 2021-01-22 16:31:27.280503

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "47b8cabac4bb"
down_revision = "6100cbbcaf7e"
branch_labels = None
depends_on = None


stmt_drop = "drop view if exists vw_region_flow_emissions;"

stmt = """
create view vw_region_flow_emissions as select
    f.trading_interval at time zone 'AEST' as trading_interval,
    date_trunc('day', f.trading_interval) at time zone 'AEST' as ti_day_aest,
    date_trunc('month', f.trading_interval) at time zone 'AEST' as ti_month_aest,
    sum(f.energy) as flow_energy,
    f.network_region || '->' || f.interconnector_region_to as flow_region,
    f.network_region as flow_from,
    f.interconnector_region_to as flow_to,
    abs(sum(ef.emissions_per_kw) * sum(f.energy)) as flow_from_emissions,
    abs(sum(et.emissions_per_kw) * sum(f.energy)) as flow_to_emissions
from mv_facility_all f
left join mv_region_emissions ef on ef.trading_interval = f.trading_interval and ef.network_region = f.network_region
left join mv_region_emissions et on et.trading_interval = f.trading_interval and et.network_region = f.interconnector_region_to
where
    f.interconnector is True
group by 1, 2, 3, flow_region, 5, 6, f.interconnector_region_to
order by trading_interval desc, flow_from asc, flow_to asc;
"""


def upgrade() -> None:
    with op.get_context().autocommit_block():
        op.execute(stmt_drop)
        op.execute(stmt)


def downgrade() -> None:
    op.execute(stmt_drop)



import logging
from datetime import datetime
from typing import Optional

from pydantic import BaseModel, validator
from pydantic.error_wrappers import ValidationError

from opennem.core.normalizers import (
    clean_float,
    normalize_aemo_region,
    normalize_duid,
)
from opennem.utils.dates import parse_date

logger = logging.getLogger(__name__)


def mms_alias_generator(name: str) -> str:
    """
    This method aliases those capitalised aemo
    names into something more sensible for python
    access
    """
    _name = name.upper()

    return _name


def validate_date(date_value: str) -> datetime:
    dt = parse_date(date_value)

    if not dt:
        raise ValueError("Not a valid date: {}".format(date_value))

    return dt


def capitalize_string(string_val: str) -> str:
    return string_val.capitalize()


class MMSBase(BaseModel):
    class Config:
        anystr_strip_whitespace = True
        use_enum_values = True
        arbitrary_types_allowed = True
        validate_assignment = True
        allow_population_by_field_name = True
        alias_generator = mms_alias_generator


# pylint: disable=no-self-argument
class ParticipantMNSPInterconnector(MMSBase):
    """"""

    linkid: str
    effectivedate: datetime
    interconnectorid: str
    fromregion: str
    toregion: str
    maxcapacity: float
    authoriseddate: datetime

    @validator("linkid", pre=True)
    def validate_linkid(cls, value: str) -> str:
        return normalize_duid(value)

    @validator("effectivedate", pre=True)
    def validate_effectivedate(cls, value: str) -> datetime:
        dt = parse_date(value)

        if not dt:
            raise ValueError("Not a valid date: {}".format(value))

        return dt

    @validator("maxcapacity", pre=True)
    def validate_maxcapacity(cls, value: str) -> float:
        f = clean_float(value)

        if not f:
            raise ValueError("Not a valid capacity: {}".format(value))

        return f

    @validator("authoriseddate", pre=True)
    def validate_authoriseddate(cls, value: str) -> datetime:
        dt = parse_date(value)

        if not dt:
            logger.error("Not a valid date: {}".format(value))
            raise ValueError("Not a valid authoriseddate: {}".format(value))

        return dt


# pylint: disable=no-self-argument
class MarketConfigInterconnector(MMSBase):
    """"""

    interconnectorid: str
    regionfrom: str
    regionto: str
    description: str
    lastchanged: datetime

    # validators
    _validate_description = validator("description")(capitalize_string)
    _validate_lastchanged = validator("lastchanged", pre=True)(validate_date)


# Map AEMO full table names to schemas
TABLE_TO_SCHEMA_MAP = {
    "PARTICIPANT_REGISTRATION_MNSP_INTERCONNECTOR": ParticipantMNSPInterconnector,
    "MARKET_CONFIG_INTERCONNECTOR": MarketConfigInterconnector,
}


def get_mms_schema_for_table(table_name: str) -> Optional[MMSBase]:
    if table_name not in TABLE_TO_SCHEMA_MAP.keys():
        logger.info("No schema found for table: {}".format(table_name))
        return None

    return TABLE_TO_SCHEMA_MAP[table_name]



import csv
import logging
from pkgutil import get_data
from typing import Dict, Optional

logger = logging.getLogger(__name__)

LEGACY_FUELTECH_MAP = {
    "brown_coal": "coal_brown",
    "black_coal": "coal_black",
    "solar": "solar_utility",
    "biomass": "bioenergy_biomass",
}


def clean_fueltech(ft: str) -> Optional[str]:
    """
    Clean the fueltech strings that come out of the spreadsheets
    or other sources

    @TODO replace with a compiled regexp
    """
    if not type(ft) is str:
        return None

    if ft == "-":
        return None

    ft = ft.lower().strip()

    if ft == "":
        return None

    return ft


def load_fueltech_map(fixture_name: str) -> Dict:
    """
    Reads the CSV to load the fueltech map

    Fields are:

    fuel_source,fuel_source_desc,tech,tech_desc,fueltech_map,load_type

    """

    MAP_KEYS = [
        "fuel_source",
        "fuel_source_desc",
        "tech",
        "tech_desc",
        "load_type",
    ]

    # Funky encoding because of save from Excel .. leave it
    csv_data = get_data("opennem", "core/data/{}".format(fixture_name))

    if not csv_data:
        raise Exception("Could not load fixture: {}".format(fixture_name))

    csv_data = csv_data.decode("utf-8-sig").splitlines()

    fueltech_map = {}

    csvreader = csv.DictReader(csv_data)
    for line in csvreader:
        record = [i or None for field, i in line.items() if field in MAP_KEYS]

        key = tuple(map(clean_fueltech, record))

        fueltech_map[key] = line["fueltech_map"]

    return fueltech_map


FUELTECH_MAP = load_fueltech_map("aemo_fueltech_map.csv")


def lookup_fueltech(
    fueltype: str,
    fueltype_desc=None,
    techtype=None,
    techtype_desc=None,
    dispatch_type: str = "generator",
) -> Optional[str]:
    """
    Takes fueltech strings from AEMO or other sources and maps them
    to opennem fueltechs using the csv file in the data directory

    """

    ft = clean_fueltech(fueltype)
    tt = clean_fueltech(techtype)
    ftd = clean_fueltech(fueltype_desc)
    ttd = clean_fueltech(techtype_desc)
    dispatch_type = dispatch_type.strip().lower()

    if dispatch_type not in ["generator", "load"]:
        raise Exception("Invalid dispatch type: {}".format(dispatch_type))

    lookup_set = tuple([ft, ftd, tt, ttd, dispatch_type])

    # Lookup legacy fuel tech types and map them
    if ft in LEGACY_FUELTECH_MAP.keys():
        return LEGACY_FUELTECH_MAP[ft]

    if lookup_set in FUELTECH_MAP:
        return FUELTECH_MAP[lookup_set]

    logger.warning("Found fueltech {}, {}, {}, {} with no mapping".format(ft, tt, ftd, ttd))

    return None



from itertools import groupby
from operator import attrgetter
from typing import Dict, List, Optional

from datetime_truncate import truncate

from opennem.api.stats.schema import DataQueryResult, RegionFlowEmissionsResult, RegionFlowResult
from opennem.api.time import human_to_interval
from opennem.schema.time import TimeInterval


def net_flows(region: str, data: List[RegionFlowResult]) -> Dict[str, List[DataQueryResult]]:
    """
    Calculates net region flows for a region from a RegionFlowResult
    """

    output_set = {}

    for k, v in groupby(data, attrgetter("interval")):
        values = list(v)

        if k not in output_set:
            output_set[k] = {
                "imports": 0.0,
                "exports": 0.0,
            }

        flow_sum = 0.0
        flow_type = "imports"

        # Sum up
        for es in values:

            if not es.generated:
                continue

            if es.flow_from == region:
                flow_sum += es.generated

            if es.flow_to == region:
                flow_sum += -1 * es.generated

        if flow_sum > 0:
            flow_type = "exports"

        output_set[k][flow_type] = flow_sum

    imports_list = []
    exports_list = []

    for interval, data in output_set.items():
        imports_list.append(
            DataQueryResult(interval=interval, group_by="imports", result=data["imports"])
        )
        exports_list.append(
            DataQueryResult(interval=interval, group_by="exports", result=data["exports"])
        )

    return {"imports": imports_list, "exports": exports_list}


def net_flows_emissions(
    region: str, data: List[RegionFlowEmissionsResult], interval: Optional[TimeInterval] = None
) -> Dict[str, List[DataQueryResult]]:
    """
    Calculates net region flow emissions for a region from a RegionFlowResult
    """

    if not interval:
        interval = human_to_interval("1d")

    output_set = {}

    for k, v in groupby(data, lambda x: truncate(x.interval, interval.trunc)):
        values = list(v)

        if k not in output_set:
            output_set[k] = {
                "imports": 0.0,
                "exports": 0.0,
            }

        export_emissions_sum = 0.0
        import_emissions_sum = 0.0

        # Sum up
        for es in values:

            if not es.flow_from:
                continue

            if es.flow_from == region:
                if es.energy > 0:
                    if es.flow_from_emissions:
                        export_emissions_sum += es.flow_from_emissions
                else:
                    if es.flow_to_emissions:
                        import_emissions_sum += es.flow_to_emissions

            if es.flow_to == region:
                if es.energy < 0:
                    if es.flow_from_emissions:
                        export_emissions_sum += es.flow_from_emissions
                else:
                    if es.flow_to_emissions:
                        import_emissions_sum += es.flow_to_emissions

        output_set[k]["imports"] = import_emissions_sum
        output_set[k]["exports"] = export_emissions_sum

    imports_list = []
    exports_list = []

    for interval, data in output_set.items():
        imports_list.append(
            DataQueryResult(interval=interval, group_by="imports", result=data["imports"])
        )
        exports_list.append(
            DataQueryResult(interval=interval, group_by="exports", result=data["exports"])
        )

    return {"imports": imports_list, "exports": exports_list}



# pylint: disable=no-member
"""
Material view NEM 5 minute facility power

Revision ID: b30e47defd26
Revises: 4569d3c0509e
Create Date: 2020-12-24 15:49:45.636162

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "b30e47defd26"
down_revision = "4569d3c0509e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
    drop materialized view if exists mv_nem_facility_power_5min cascade;

    CREATE MATERIALIZED VIEW mv_nem_facility_power_5min WITH (timescaledb.continuous) AS
    select
        time_bucket('5 minutes', fs.trading_interval) as trading_interval,
        fs.facility_code,
        max(fs.generated) as facility_power
    from facility_scada fs
    where fs.network_id = 'NEM'
    group by
        1,
        fs.facility_code;

    """
    )


def downgrade() -> None:
    op.execute("drop view mv_nem_facility_power_5min")



# pylint: disable=no-member
"""Network additional timezone_database and offset fields

Revision ID: 3595c70244db
Revises: f04dfaf47b51
Create Date: 2020-10-09 01:05:10.156455

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3595c70244db"
down_revision = "f04dfaf47b51"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("network", sa.Column("offset", sa.Integer(), nullable=True))
    op.add_column(
        "network", sa.Column("timezone_database", sa.Text(), nullable=True)
    )


def downgrade():
    op.drop_column("network", "timezone_database")
    op.drop_column("network", "offset")



import json
import logging
from typing import List

from pydantic import ValidationError

from opennem.importer.all import run_all
from opennem.importer.opennem import opennem_import
from opennem.schema.opennem import StationSchema

logger = logging.getLogger("opennem.exporter")


def load_stations(file_path: str = "opennem.json") -> List[StationSchema]:
    with open("data/opennem.json") as fh:
        __data = json.load(fh)

    stations: List[StationSchema] = []

    for i in __data.values():
        try:
            stations.append(StationSchema(**i))
        except ValidationError as e:
            logger.error(
                "Error with record: {} {}: {}".format(i["code"], i["name"], e)
            )

    return stations


if __name__ == "__main__":
    run_all()
    stations = opennem_import()



# pylint: disable=no-member
"""
Fix energy quantities for NEM

Revision ID: df7308f58ee2
Revises: d886acb494ca
Create Date: 2020-12-02 01:53:38.801723

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "df7308f58ee2"
down_revision = "d886acb494ca"
branch_labels = None
depends_on = None


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass



from datetime import datetime, timedelta
from textwrap import dedent
from typing import List, Optional

from sqlalchemy import sql
from sqlalchemy.sql.elements import TextClause

from opennem.api.stats.controllers import get_scada_range, networks_to_in
from opennem.api.stats.schema import ScadaDateRange
from opennem.schema.network import NetworkSchema, NetworkWEM
from opennem.schema.stats import StatTypes
from opennem.schema.time import TimeInterval, TimePeriod
from opennem.utils.dates import subtract_week


def interconnector_flow_emissions_query(
    network_region: str, date_range: ScadaDateRange, network: NetworkSchema
) -> str:
    """Get emissions for flows for a region"""
    __query = """
        select
            f.trading_interval at time zone '{timezone}' as trading_interval,
            sum(f.energy) as flow_energy,
            f.network_region || '->' || f.interconnector_region_to as flow_region,
            f.network_region as flow_from,
            f.interconnector_region_to as flow_to,
            abs(sum(ef.emissions_per_kw) * sum(f.energy)) as flow_from_emissions,
            abs(sum(et.emissions_per_kw) * sum(f.energy)) as flow_to_emissions
        from mv_facility_all f
        left join mv_region_emissions ef on ef.trading_interval = f.trading_interval and ef.network_region = f.network_region
        left join mv_region_emissions et on et.trading_interval = f.trading_interval and et.network_region = f.interconnector_region_to
        where
            f.interconnector is True and
            (f.network_region= '{region}' or f.interconnector_region_to= '{region}') and
            fs.trading_interval <= '{date_end}' and
            fs.trading_interval > '{date_start}'
        group by 1, flow_region, 3, 4, f.interconnector_region_to
        order by trading_interval desc, flow_from asc, flow_to asc;
    """.format(
        timezone=network.timezone_database,
        region=network_region,
        date_start=subtract_week(date_range.get_end()),
        date_end=date_range.get_end(),
    )

    return dedent(__query)


def interconnector_flow_power_query(
    network_region: str, date_range: ScadaDateRange, network: NetworkSchema
) -> str:
    __query = """
            select
                -- time_bucket_gapfill('5 minutes', trading_interval) as trading_interval,
                trading_interval at time zone '{timezone}' as trading_interval,
                max(fs.facility_power) as facility_power,
                f.network_region as flow_from,
                f.interconnector_region_to as flow_to
            from mv_nem_facility_power_5min fs
            join facility f on fs.facility_code = f.code
            where
                f.interconnector is True and
                (f.network_region= '{region}' or f.interconnector_region_to= '{region}') and
                fs.trading_interval <= '{date_end}' and
                fs.trading_interval > '{date_start}'
            group by 1, f.code, 3, 4
            order by trading_interval desc;
           """.format(
        timezone=network.timezone_database,
        region=network_region,
        date_start=subtract_week(date_range.get_end()),
        date_end=date_range.get_end(),
    )

    return dedent(__query)


def interconnector_power_flow(
    network_region: str, date_range: ScadaDateRange, network: NetworkSchema
) -> str:
    """Get region flows using materialized view"""
    ___query = """
        select
            trading_interval at time zone '{timezone}' as trading_interval,
            fs.network_region,
            fs.imports,
            fs.exports
        from mv_interchange_power_nem_region fs
        where
            fs.network_id = '{network_id}' and
            fs.network_region= '{region}' and
            fs.trading_interval <= '{date_end}' and
            fs.trading_interval > '{date_start}'
        order by trading_interval desc;
    """.format(
        timezone=network.timezone_database,
        network_id=network.code,
        region=network_region,
        date_start=subtract_week(date_range.get_end()),
        date_end=date_range.get_end(),
    )

    return dedent(___query)


def interconnector_net_energy_flow(network_region: str, date_range: ScadaDateRange) -> TextClause:
    """Get interconnector energy flows for a region"""
    __query = sql.text(
        dedent(
            """
            select
                f.trading_interval,
                max(f.energy) as facility_energy,
                f.network_region as flow_from,
                f.interconnector_region_to as flow_to
            from mv_facility_all f
            where
                f.interconnector is True and
                (f.network_region= :region or f.interconnector_region_to= :region) and
                f.trading_interval <= :date_end and
                f.trading_interval > :date_start
            group by 1, 3, 4
           """
        )
    ).bindparams(
        region=network_region,
        date_start=date_range.get_start(),
        date_end=date_range.get_end(),
    )

    return __query


def country_stats_query(stat_type: StatTypes, country: str = "au") -> TextClause:
    __query = sql.text(
        dedent(
            """
                select
                    s.stat_date,
                    s.value,
                    s.stat_type
                from stats s
                where s.stat_type = :stat_type and s.country= :country
                order by s.stat_date desc
           """
        )
    ).bindparams(
        stat_type=str(stat_type),
        country=country,
    )

    return __query


def price_network_query(
    interval: TimeInterval,
    network: NetworkSchema = NetworkWEM,
    group_field: str = "bs.network_id",
    date_range: Optional[ScadaDateRange] = None,
    period: Optional[TimePeriod] = None,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:

    if not networks_query:
        networks_query = [network]

    if network not in networks_query:
        networks_query.append(network)

    if not date_range:
        date_range = get_scada_range(network=network, networks=networks_query)

    # timezone = network.get_timezone(postgres_format=True)
    timezone = network.timezone_database

    __query = """
        select
            trading_interval at time zone '{timezone}',
            {group_field},
            avg(bs.price) as price
        from balancing_summary bs
        where
            bs.trading_interval <= '{date_max}' and
            bs.trading_interval >= '{date_min}' and
            {network_query}
            {network_region_query}
            1=1
        group by 1, 2
        order by 1 desc
    """

    network_region_query = ""

    if network_region:
        network_region_query = f"bs.network_region='{network_region}' and "
        group_field = "bs.network_region"

    network_query = "bs.network_id IN ({}) and ".format(networks_to_in(networks_query))

    if len(networks_query) > 1:
        group_field = "'AU'"

    if not date_range:
        raise Exception("Require a date range")

    date_max = date_range.get_end()
    date_min = date_range.get_start()

    if period:
        date_min = date_range.get_end() - timedelta(minutes=period.period)

    query = dedent(
        __query.format(
            network_query=network_query,
            trunc=interval.interval_sql,
            network_region_query=network_region_query,
            timezone=timezone,
            date_max=date_max,
            date_min=date_min,
            group_field=group_field,
        )
    )

    return query


def power_network_fueltech_query(
    interval: TimeInterval,
    network: NetworkSchema = NetworkWEM,
    date_range: Optional[ScadaDateRange] = None,
    period: Optional[TimePeriod] = None,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> str:
    """Query power stats"""

    if not networks_query:
        networks_query = [network]

    if network not in networks_query:
        networks_query.append(network)

    if not date_range:
        date_range = get_scada_range(network=network, networks=networks_query)

    table_query = "mv_nem_facility_power_5min"

    if network == NetworkWEM:
        table_query = "mv_wem_facility_power_30min"

    __query = """
        select
            fs.trading_interval at time zone '{timezone}' AS trading_interval,
            ft.code as fueltech_code,
            sum(fs.facility_power) as facility_power
        from {table_query} fs
        join facility f on fs.facility_code = f.code
        join fueltech ft on f.fueltech_id = ft.code
        where
            f.fueltech_id is not null and
            f.fueltech_id not in ('exports', 'imports') and
            {network_query}
            {network_region_query}
            fs.trading_interval <= '{date_max}' and
            fs.trading_interval > '{date_min}'
            {fueltech_filter}
        group by 1, 2
    """

    network_region_query: str = ""
    fueltech_filter: str = ""
    wem_apvi_case: str = ""
    timezone: str = network.timezone_database

    if network_region:
        network_region_query = f"f.network_region='{network_region}' and "
    else:
        fueltech_filter = "and f.fueltech_id not in ('imports', 'exports') "

    if NetworkWEM in networks_query:
        # silly single case we'll refactor out
        # APVI network is used to provide rooftop for WEM so we require it
        # in country-wide totals
        wem_apvi_case = "or (f.network_id='APVI' and f.network_region='WEM')"

    network_query = "(f.network_id IN ({}) {}) and ".format(
        networks_to_in(networks_query), wem_apvi_case
    )

    if not date_range:
        raise Exception("Require a date range")

    date_max = date_range.get_end()
    date_min = date_range.get_start()

    if period:
        date_min = date_range.get_end() - timedelta(minutes=period.period)

    query = dedent(
        __query.format(
            table_query=table_query,
            network_query=network_query,
            trunc=interval.interval_sql,
            network_region_query=network_region_query,
            timezone=timezone,
            date_max=date_max,
            date_min=date_min,
            fueltech_filter=fueltech_filter,
            wem_apvi_case=wem_apvi_case,
        )
    )

    return query


def energy_network_fueltech_query(
    network: NetworkSchema,
    interconnector: bool = False,
    emissions: bool = False,
    interval: Optional[TimeInterval] = None,
    year: Optional[int] = None,
    network_region: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
    date_range: Optional[ScadaDateRange] = None,
) -> str:
    """
    Get Energy for a network or network + region
    based on a year
    """

    if not networks_query:
        networks_query = [network]

    if network not in networks_query:
        networks_query.append(network)

    __query = """
    select
        t.ti_{trunc_name} as trading_day,
        t.fueltech_id,
        sum(t.energy) / 1000 as fueltech_energy,
        sum(t.market_value) as fueltech_market_value,
        sum(t.emissions) as fueltech_emissions
    from mv_facility_all t
    where
        t.trading_interval <= '{year_max}' and
        t.trading_interval >= '{year_min}' and
        {network_query}
        {network_region_query}
        {fueltech_filter}
        1=1
    group by 1, 2
    order by
        trading_day desc;
    """

    if interconnector:
        # __query = """
        # select
        #     t.ti_{trunc_name} as trading_day,
        #     sum(t.energy) / 1000 as interconnector_energy,
        #     t.network_region as flow_from,
        #     t.interconnector_region_to as flow_to,
        #     t.network_region || '->' || t.interconnector_region_to as flow_region
        # from mv_facility_all t
        # where
        #     t.interconnector is True and
        #     t.trading_interval <= '{year_max}' and
        #     t.trading_interval >= '{year_min}' and
        #     {network_query}
        #     {network_region_query}
        #     {fueltech_filter}
        #     1=1
        # group by 1, 5, 3, 4
        # order by
        #     trading_day desc;
        # """

        __query = """
        select
            date_trunc('{trunc}', t.trading_interval at time zone '{timezone}') as trading_day,
            sum(t.imports_avg) / 1000,
            sum(t.exports_avg) / 1000
        from mv_interchange_energy_nem_region t
        where
            t.trading_interval <= '{year_max}' and
            t.trading_interval >= '{year_min}' and
            {network_query}
            {network_region_query}
            1=1
        group by 1
        order by 1 desc

        """

    if emissions:
        __query = """
        select
            t.trading_interval at time zone '{timezone}' as trading_interval,
            t.flow_from,
            t.flow_to,
            sum(t.flow_energy) as energy,
            max(t.flow_from_emissions),
            max(t.flow_to_emissions)
        from vw_region_flow_emissions t
        where
            t.trading_interval <= '{year_max}' and
            t.trading_interval >= '{year_min}' and
            {network_region_query}
            1=1
        group by 1, t.flow_region, 2, 3
        order by 1 desc

        """

    timezone = network.timezone_database
    offset = network.get_timezone(postgres_format=True)

    if not date_range:
        date_range = get_scada_range(network=network, networks=networks_query)

    if not date_range:
        raise Exception("Require a scada range for {}".format(network.code))

    network_region_query = ""
    fueltech_filter = ""

    if network_region:
        # network_region_interconnector = ""

        # if interconnector:
        # network_region_interconnector = f" or t.interconnector_region_to='{network_region}'"

        network_region_query = f"t.network_region='{network_region}' and"

    fueltech_filter = "t.fueltech_id not in ('imports', 'exports') and "

    if emissions:
        network_region_query = f"""
            (t.flow_from='{network_region}' or t.flow_to='{network_region}') and
        """

    # if interconnector:
    # fueltech_filter = "t.fueltech_id in ('imports', 'exports') and "

    networks_list = networks_to_in(networks_query)
    network_query = "t.network_id IN ({}) and ".format(networks_list)

    if not timezone:
        timezone = "UTC"

    trunc = None

    if interval:
        trunc = interval.trunc

    if year:
        # might have to do +offset times
        year_max = "{}-12-31 23:59:59{}".format(year, offset)
        year_min = "{}-01-01 00:00:00{}".format(year, offset)

        if year == datetime.now().year:
            year_max = date_range.get_end()

        if not trunc:
            trunc = "day"
    else:
        year_min = date_range.get_start()
        year_max = date_range.get_end()

        if not trunc:
            trunc = "month"

    trunc_name = "{}_{}".format(trunc, timezone).lower()

    query = dedent(
        __query.format(
            timezone=network.timezone_database,
            trunc=trunc,
            trunc_name=trunc_name,
            year_min=year_min,
            year_max=year_max,
            network_query=network_query,
            network_region_query=network_region_query,
            fueltech_filter=fueltech_filter,
        )
    )

    return query



# pylint: disable=no-member
"""
update indicies that aren't used

Revision ID: 4bf86ff5c8ff
Revises: 64987ea01b57
Create Date: 2020-11-23 02:54:29.564574

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "4bf86ff5c8ff"
down_revision = "64987ea01b57"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_index("idx_facility_scada_trading_interval_perth_year")
    op.drop_index("idx_facility_scada_trading_interval_perth_month")
    op.drop_index("idx_facility_scada_trading_interval_perth_day")
    op.drop_index("idx_facility_scada_trading_interval_perth_hour")

    op.drop_index("idx_balancing_summary_trading_interval_perth_year")
    op.drop_index("idx_balancing_summary_trading_interval_perth_month")
    op.drop_index("idx_balancing_summary_trading_interval_perth_day")
    op.drop_index("idx_balancing_summary_trading_interval_perth_hour")

    op.drop_index("idx_facility_scada_trading_interval_sydney_year")
    op.drop_index("idx_facility_scada_trading_interval_sydney_month")
    op.drop_index("idx_facility_scada_trading_interval_sydney_day")
    op.drop_index("idx_facility_scada_trading_interval_sydney_hour")

    op.drop_index("idx_balancing_summary_trading_interval_sydney_year")
    op.drop_index("idx_balancing_summary_trading_interval_sydney_month")
    op.drop_index("idx_balancing_summary_trading_interval_sydney_day")
    op.drop_index("idx_balancing_summary_trading_interval_sydney_hour")


def downgrade():
    pass



"""
Import and parse CPI stats
"""

import logging
from typing import List

from opennem.core.parsers.cpi import stat_au_cpi
from opennem.core.stats.store import store_stats_database

if __name__ == "__main__":
    r = stat_au_cpi()
    store_stats_database(r)



# pylint: disable=no-member
"""BOMStation additional feed fields

Revision ID: f04dfaf47b51
Revises: a76e6ff50cda
Create Date: 2020-10-08 01:04:58.796972

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f04dfaf47b51"
down_revision = "a76e6ff50cda"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "bom_station", sa.Column("feed_url", sa.Text(), nullable=True)
    )
    op.add_column(
        "bom_station", sa.Column("is_capital", sa.Boolean(), nullable=True)
    )
    op.add_column(
        "bom_station", sa.Column("name_alias", sa.Text(), nullable=True)
    )
    op.add_column(
        "bom_station", sa.Column("priority", sa.Integer(), nullable=True)
    )
    op.add_column(
        "bom_station", sa.Column("website_url", sa.Text(), nullable=True)
    )


def downgrade():
    op.drop_column("bom_station", "website_url")
    op.drop_column("bom_station", "priority")
    op.drop_column("bom_station", "name_alias")
    op.drop_column("bom_station", "is_capital")
    op.drop_column("bom_station", "feed_url")



import logging
from datetime import datetime, timedelta

from opennem.core.networks import network_from_network_code
from opennem.notifications.slack import slack_message
from opennem.settings import settings
from opennem.utils.dates import chop_microseconds, parse_date
from opennem.utils.http import http

logger = logging.getLogger("opennem.monitors.opennem")


def check_opennem_interval_delays(network_code: str) -> bool:
    network = network_from_network_code(network_code)
    url = (
        f"https://data.opennem.org.au/v3/stats/au/{network.code}/power/7d.json"
    )

    resp = http.get(url)

    if resp.status_code != 200 or not resp.ok:
        logger.error("Error retrieving: {}".format(url))
        return False

    resp_json = resp.json()

    if "data" not in resp_json:
        logger.error("Error retrieving wem power: malformed response")
        return False

    data = resp_json["data"]

    fueltech_data = data.pop()

    history_end_date = fueltech_data["history"]["last"]

    history_date = parse_date(history_end_date, dayfirst=False)

    if not history_date:
        logger.error(
            "Could not read history date for opennem interval monitor"
        )
        return False

    now_date = datetime.now().astimezone(network.get_timezone())

    time_delta = chop_microseconds(now_date - history_date)

    logger.debug("Live time: {},  delay: {}".format(history_date, time_delta))

    if time_delta > timedelta(hours=3):
        slack_message(
            "*WARNING*: OpenNEM {} interval delay on {} currently: {}\n".format(
                network.code, settings.env, time_delta
            )
        )

    return True


if __name__ == "__main__":
    for network_code in ["NEM", "WEM"]:
        check_opennem_interval_delays(network_code)



# pylint: disable=no-member
"""
create mv_interchange_energy_nem_region

Revision ID: 26ecffa41c19
Revises: 4c35e2483eb7
Create Date: 2021-01-22 01:18:57.403176

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "26ecffa41c19"
down_revision = "4c35e2483eb7"
branch_labels = None
depends_on = None

stmt_drop = "drop materialized view if exists mv_interchange_energy_nem_region cascade;"

stmt = """
    CREATE MATERIALIZED VIEW mv_interchange_energy_nem_region WITH (timescaledb.continuous) AS
    select
        time_bucket('1 hour', bs.trading_interval) as trading_interval,
        bs.network_id,
        bs.network_region,
        case when sum(bs.net_interchange) < 0 then
            sum(bs.net_interchange) / 12
        else 0
        end as imports_avg,
        case when sum(bs.net_interchange) > 0 then
            sum(bs.net_interchange) / 12
        else 0
        end as exports_avg,
        --    energy sums
        case when sum(bs.net_interchange) < 0 then
            energy_sum(bs.net_interchange, '1 hour') * interval_size('1 hour', count(bs.net_interchange))
        else 0
        end as imports_energy,
        case when sum(bs.net_interchange) > 0 then
            energy_sum(bs.net_interchange, '1 hour') * interval_size('1 hour', count(bs.net_interchange))
        else 0
        end as exports_energy
    from balancing_summary bs
    group by 1, 2, 3;
"""


def upgrade() -> None:
    with op.get_context().autocommit_block():
        op.execute(stmt_drop)
        op.execute(stmt)


def downgrade() -> None:
    op.execute(stmt_drop)



import os
import sys
from pathlib import Path


def running_as_scrapy() -> bool:
    """
    Detect if we're running as scrapy
    """

    # running in scrapyd
    if "SCRAPYD" in os.environ:
        return True

    # running as lib
    if not sys.argv or len(sys.argv) < 1:
        return False

    _cl = sys.argv[0]

    if Path(_cl).name in ["scrapy"]:
        return True

    return False



# pylint: disable=no-member
"""This is the base migration as of this date. Don't touch this.

Revision ID: 42cdf37a6e3b
Revises:
Create Date: 2020-10-06 13:53:50.693303

"""
import geoalchemy2
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

revision = "42cdf37a6e3b"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.execute("CREATE EXTENSION IF NOT EXISTS postgis")

    op.create_table(
        "bom_station",
        sa.Column("code", sa.Text(), nullable=False),
        sa.Column("state", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("registered", sa.Date(), nullable=True),
        sa.Column("lat", sa.Numeric(), nullable=True),
        sa.Column("lng", sa.Numeric(), nullable=True),
        sa.PrimaryKeyConstraint("code"),
    )
    op.create_table(
        "facility_status",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("code", sa.Text(), nullable=False),
        sa.Column("label", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("code"),
    )
    op.create_table(
        "fueltech",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("code", sa.Text(), nullable=False),
        sa.Column("label", sa.Text(), nullable=True),
        sa.Column("renewable", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("code"),
    )
    op.create_table(
        "location",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("address1", sa.Text(), nullable=True),
        sa.Column("address2", sa.Text(), nullable=True),
        sa.Column("locality", sa.Text(), nullable=True),
        sa.Column("state", sa.Text(), nullable=True),
        sa.Column("postcode", sa.Text(), nullable=True),
        sa.Column("place_id", sa.Text(), nullable=True),
        sa.Column("geocode_approved", sa.Boolean(), nullable=True),
        sa.Column("geocode_skip", sa.Boolean(), nullable=True),
        sa.Column("geocode_processed_at", sa.DateTime(), nullable=True),
        sa.Column("geocode_by", sa.Text(), nullable=True),
        sa.Column(
            "geom",
            geoalchemy2.types.Geometry(
                geometry_type="POINT",
                srid=4326,
                spatial_index=False,
                from_text="ST_GeomFromEWKT",
                name="geometry",
            ),
            nullable=True,
        ),
        sa.Column(
            "boundary",
            geoalchemy2.types.Geometry(
                geometry_type="MULTIPOLYGON",
                srid=4326,
                spatial_index=False,
                from_text="ST_GeomFromEWKT",
                name="geometry",
            ),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "idx_location_boundary",
        "location",
        ["boundary"],
        unique=False,
        postgresql_using="gist",
    )
    op.create_index(
        "idx_location_geom",
        "location",
        ["geom"],
        unique=False,
        postgresql_using="gist",
    )
    op.create_index(
        op.f("ix_location_place_id"), "location", ["place_id"], unique=False
    )
    op.create_table(
        "network",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("code", sa.Text(), nullable=False),
        sa.Column("country", sa.Text(), nullable=False),
        sa.Column("label", sa.Text(), nullable=True),
        sa.Column("timezone", sa.Text(), nullable=False),
        sa.Column("interval_size", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("code"),
    )
    op.create_table(
        "participant",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("code", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("network_name", sa.Text(), nullable=True),
        sa.Column("network_code", sa.Text(), nullable=True),
        sa.Column("country", sa.Text(), nullable=True),
        sa.Column("abn", sa.Text(), nullable=True),
        sa.Column("approved", sa.Boolean(), nullable=True),
        sa.Column("approved_by", sa.Text(), nullable=True),
        sa.Column("approved_at", sa.DateTime(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_participant_code"), "participant", ["code"], unique=True
    )
    op.create_table(
        "balancing_summary",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("network_id", sa.Text(), nullable=False),
        sa.Column(
            "trading_interval",
            postgresql.TIMESTAMP(timezone=True),
            nullable=False,
        ),
        sa.Column("forecast_load", sa.Numeric(), nullable=True),
        sa.Column("generation_scheduled", sa.Numeric(), nullable=True),
        sa.Column("generation_non_scheduled", sa.Numeric(), nullable=True),
        sa.Column("generation_total", sa.Numeric(), nullable=True),
        sa.Column("price", sa.Numeric(), nullable=True),
        sa.ForeignKeyConstraint(
            ["network_id"],
            ["network.code"],
            name="fk_balancing_summary_network_code",
        ),
        sa.PrimaryKeyConstraint("network_id", "trading_interval"),
    )
    op.create_index(
        op.f("ix_balancing_summary_trading_interval"),
        "balancing_summary",
        ["trading_interval"],
        unique=False,
    )
    op.create_table(
        "bom_observation",
        sa.Column("observation_time", sa.DateTime(), nullable=False),
        sa.Column("station_id", sa.Text(), nullable=False),
        sa.Column("temp_apparent", sa.Numeric(), nullable=True),
        sa.Column("temp_air", sa.Numeric(), nullable=True),
        sa.Column("press_qnh", sa.Numeric(), nullable=True),
        sa.Column("wind_dir", sa.Text(), nullable=True),
        sa.Column("wind_spd", sa.Numeric(), nullable=True),
        sa.ForeignKeyConstraint(
            ["station_id"],
            ["bom_station.code"],
            name="fk_bom_observation_station_code",
        ),
        sa.PrimaryKeyConstraint("observation_time", "station_id"),
    )
    op.create_table(
        "facility_scada",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("network_id", sa.Text(), nullable=False),
        sa.Column(
            "trading_interval",
            postgresql.TIMESTAMP(timezone=True),
            nullable=False,
        ),
        sa.Column("facility_code", sa.Text(), nullable=False),
        sa.Column("generated", sa.Numeric(), nullable=True),
        sa.Column("eoi_quantity", sa.Numeric(), nullable=True),
        sa.ForeignKeyConstraint(
            ["network_id"],
            ["network.code"],
            name="fk_balancing_summary_network_code",
        ),
        sa.PrimaryKeyConstraint(
            "network_id", "trading_interval", "facility_code"
        ),
    )
    op.create_index(
        op.f("ix_facility_scada_facility_code"),
        "facility_scada",
        ["facility_code"],
        unique=False,
    )
    op.create_index(
        op.f("ix_facility_scada_trading_interval"),
        "facility_scada",
        ["trading_interval"],
        unique=False,
    )
    op.create_table(
        "station",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("participant_id", sa.Integer(), nullable=True),
        sa.Column("location_id", sa.Integer(), nullable=True),
        sa.Column("code", sa.Text(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("wikipedia_link", sa.Text(), nullable=True),
        sa.Column("wikidata_id", sa.Text(), nullable=True),
        sa.Column("network_code", sa.Text(), nullable=True),
        sa.Column("network_name", sa.Text(), nullable=True),
        sa.Column("approved", sa.Boolean(), nullable=True),
        sa.Column("approved_by", sa.Text(), nullable=True),
        sa.Column("approved_at", sa.DateTime(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["location_id"], ["location.id"], name="fk_station_location_id"
        ),
        sa.ForeignKeyConstraint(
            ["participant_id"],
            ["participant.id"],
            name="fk_station_participant_id",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_station_code"), "station", ["code"], unique=False)
    op.create_index(
        op.f("ix_station_network_code"),
        "station",
        ["network_code"],
        unique=False,
    )
    op.create_table(
        "facility",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("network_id", sa.Text(), nullable=False),
        sa.Column("fueltech_id", sa.Text(), nullable=True),
        sa.Column("status_id", sa.Text(), nullable=True),
        sa.Column("station_id", sa.Integer(), nullable=True),
        sa.Column("code", sa.Text(), nullable=True),
        sa.Column("network_code", sa.Text(), nullable=True),
        sa.Column("network_region", sa.Text(), nullable=True),
        sa.Column("network_name", sa.Text(), nullable=True),
        sa.Column("active", sa.Boolean(), nullable=True),
        sa.Column(
            "dispatch_type",
            sa.Enum("GENERATOR", "LOAD", name="dispatchtype"),
            nullable=False,
        ),
        sa.Column("capacity_registered", sa.Numeric(), nullable=True),
        sa.Column("registered", sa.DateTime(), nullable=True),
        sa.Column("deregistered", sa.DateTime(), nullable=True),
        sa.Column("unit_id", sa.Integer(), nullable=True),
        sa.Column("unit_number", sa.Integer(), nullable=True),
        sa.Column("unit_alias", sa.Text(), nullable=True),
        sa.Column("unit_capacity", sa.Numeric(), nullable=True),
        sa.Column("approved", sa.Boolean(), nullable=True),
        sa.Column("approved_by", sa.Text(), nullable=True),
        sa.Column("approved_at", sa.DateTime(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["fueltech_id"], ["fueltech.code"], name="fk_facility_fueltech_id"
        ),
        sa.ForeignKeyConstraint(
            ["network_id"], ["network.code"], name="fk_station_network_code"
        ),
        sa.ForeignKeyConstraint(
            ["station_id"], ["station.id"], name="fk_station_status_code"
        ),
        sa.ForeignKeyConstraint(
            ["status_id"],
            ["facility_status.code"],
            name="fk_facility_status_code",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_facility_code"), "facility", ["code"], unique=False
    )
    op.create_index(
        op.f("ix_facility_network_code"),
        "facility",
        ["network_code"],
        unique=False,
    )
    op.create_index(
        op.f("ix_facility_network_region"),
        "facility",
        ["network_region"],
        unique=False,
    )
    op.create_table(
        "photo",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("station_id", sa.Integer(), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("mime_type", sa.Text(), nullable=True),
        sa.Column("original_url", sa.Text(), nullable=True),
        sa.Column("data", sa.LargeBinary(), nullable=True),
        sa.Column("width", sa.Integer(), nullable=True),
        sa.Column("height", sa.Integer(), nullable=True),
        sa.Column("license_type", sa.Text(), nullable=True),
        sa.Column("license_link", sa.Text(), nullable=True),
        sa.Column("author", sa.Text(), nullable=True),
        sa.Column("author_link", sa.Text(), nullable=True),
        sa.Column("processed", sa.Boolean(), nullable=True),
        sa.Column("processed_by", sa.Text(), nullable=True),
        sa.Column("processed_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("approved", sa.Boolean(), nullable=True),
        sa.Column("approved_by", sa.Text(), nullable=True),
        sa.Column("approved_at", sa.DateTime(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["station_id"], ["station.id"], name="fk_photos_station_id"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "revisions",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("station_id", sa.Integer(), nullable=True),
        sa.Column("facility_id", sa.Integer(), nullable=True),
        sa.Column("location_id", sa.Integer(), nullable=True),
        sa.Column("changes", sa.JSON(), nullable=True),
        sa.Column("previous", sa.JSON(), nullable=True),
        sa.Column("is_update", sa.Boolean(), nullable=True),
        sa.Column("approved", sa.Boolean(), nullable=True),
        sa.Column("approved_by", sa.Text(), nullable=True),
        sa.Column("approved_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("approved_comment", sa.Text(), nullable=True),
        sa.Column("discarded", sa.Boolean(), nullable=True),
        sa.Column("discarded_by", sa.Text(), nullable=True),
        sa.Column("discarded_at", sa.DateTime(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(
            ["facility_id"], ["facility.id"], name="fk_revision_facility_id"
        ),
        sa.ForeignKeyConstraint(
            ["location_id"], ["location.id"], name="fk_revision_location_id"
        ),
        sa.ForeignKeyConstraint(
            ["station_id"], ["station.id"], name="fk_revision_station_id"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("revisions")
    op.drop_table("photo")
    op.drop_index(op.f("ix_facility_network_region"), table_name="facility")
    op.drop_index(op.f("ix_facility_network_code"), table_name="facility")
    op.drop_index(op.f("ix_facility_code"), table_name="facility")
    op.drop_table("facility")
    op.drop_index(op.f("ix_station_network_code"), table_name="station")
    op.drop_index(op.f("ix_station_code"), table_name="station")
    op.drop_table("station")
    op.drop_index(
        op.f("ix_facility_scada_trading_interval"), table_name="facility_scada"
    )
    op.drop_index(
        op.f("ix_facility_scada_facility_code"), table_name="facility_scada"
    )
    op.drop_table("facility_scada")
    op.drop_table("bom_observation")
    op.drop_index(
        op.f("ix_balancing_summary_trading_interval"),
        table_name="balancing_summary",
    )
    op.drop_table("balancing_summary")
    op.drop_index(op.f("ix_participant_code"), table_name="participant")
    op.drop_table("participant")
    op.drop_table("network")
    op.drop_index(op.f("ix_location_place_id"), table_name="location")
    op.drop_index("idx_location_geom", table_name="location")
    op.drop_index("idx_location_boundary", table_name="location")
    op.drop_table("location")
    op.drop_table("fueltech")
    op.drop_table("facility_status")
    op.drop_table("bom_station")
    # ### end Alembic commands ###



# pylint: disable=no-member
"""
Make energy_sum as parallel safe

Revision ID: d8bd4db7c53d
Revises: e79ef4fd3653
Create Date: 2020-12-11 17:13:27.201643

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "d8bd4db7c53d"
down_revision = "e79ef4fd3653"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
        create or replace function energy_sum_combine_states(agg_state agg_state_energy, agg_state2 agg_state_energy)
        returns agg_state_energy
        immutable
        language plpgsql
        as $$
        declare
            agg_state_return agg_state_energy;

        begin
            agg_state_return = agg_state;
            agg_state_return.energy_sum = agg_state_return.energy_sum + agg_state2.energy_sum;

            return agg_state_return;
        end;
        $$;

        create or replace aggregate energy_sum (numeric, bucket_size text)
        (
            sfunc = sum_area,
            stype = agg_state_energy,
            finalfunc = energy_sum_pop,
            initcond = '(0,,0.0, TRUE)',
            COMBINEFUNC = energy_sum_combine_states,
            PARALLEL = SAFE
        );
        """
    )


def downgrade() -> None:
    pass



from typing import Optional

from opennem.core.loader import load_data


def load_fueltechs() -> dict:
    fueltechs = load_data("fueltechs.json", from_fixture=True)

    fueltechs_dict = {}

    for s in fueltechs:
        _code = s.get("code", None)
        fueltechs_dict[_code] = s

    return fueltechs_dict


FACILITY_FUELTECH_FIXTURE = load_fueltechs()


def parse_facility_fueltech(fueltech_code: Optional[str]) -> Optional[dict]:
    if fueltech_code is None:
        return None

    if fueltech_code not in FACILITY_FUELTECH_FIXTURE.keys():
        raise Exception("Invalid facility fueltech: {}".format(fueltech_code))

    return FACILITY_FUELTECH_FIXTURE[fueltech_code]



from typing import List

from fastapi import APIRouter, Depends, HTTPException, Query
from sqlalchemy.orm import Session

from opennem.db import get_database_session
from opennem.db.models.opennem import Location
from opennem.schema.opennem import LocationSchema

router = APIRouter()


@router.get("/", description="Locations", response_model=List[LocationSchema])
def locations(
    session: Session = Depends(get_database_session),
) -> List[LocationSchema]:
    locations = session.query(Location).all()

    return locations



# pylint: disable=no-member
"""
Update old rooftop stations

Revision ID: 15f5aaf2699c
Revises: 0716d393ff9d
Create Date: 2020-11-17 15:05:40.037459

"""
import geoalchemy2
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "15f5aaf2699c"
down_revision = "0716d393ff9d"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "update facility set code = 'ROOFTOP_APVI_TAS' where code = 'ROOFTOP_TAS'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_APVI_QLD' where code = 'ROOFTOP_QLD'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_APVI_NSW' where code = 'ROOFTOP_NSW'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_APVI_VIC' where code = 'ROOFTOP_VIC'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_APVI_SA' where code = 'ROOFTOP_SA'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_APVI_WA' where code = 'ROOFTOP_WA'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_APVI_NT' where code = 'ROOFTOP_NT'"
    )


def downgrade():
    op.execute(
        "update facility set code = 'ROOFTOP_TAS' where code = 'ROOFTOP_APVI_TAS'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_QLD' where code = 'ROOFTOP_APVI_QLD'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_VIC' where code = 'ROOFTOP_APVI_VIC'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_SA' where code = 'ROOFTOP_APVI_SA'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_WA' where code = 'ROOFTOP_APVI_WA'"
    )
    op.execute(
        "update facility set code = 'ROOFTOP_NT' where code = 'ROOFTOP_APVI_NT'"
    )



from opennem.schema.network import (
    NETWORKS,
    NetworkAPVI,
    NetworkAU,
    NetworkNEM,
    NetworkSchema,
    NetworkWEM,
)

NEM_STATES = ["QLD", "NSW", "VIC", "ACT", "TAS", "SA", "NT"]


def state_from_network_region(network_region: str) -> str:
    _state = network_region

    if _state.endswith("1"):
        _state = _state[:-1]

    _state = _state.strip().upper()

    if _state in NEM_STATES:
        return _state

    raise Exception("State {} not found".format(network_region))


def network_from_state(state: str) -> NetworkSchema:
    state = state.upper().strip()

    if state in ["WA"]:
        return NetworkWEM

    if state in ["QLD", "NSW", "VIC", "ACT", "TAS", "SA", "NT"]:
        return NetworkNEM

    raise Exception("Unknown network {}".format(state))


def network_from_network_region(
    network_region: str,
) -> NetworkSchema:
    network_region = network_region.upper()

    if network_region in ["WEM", "WA1"]:
        return NetworkWEM
    if network_region in ["NEM", "NSW1", "QLD1", "SA1", "VIC1", "TAS1"]:
        return NetworkNEM

    raise Exception("Unknown network {}".format(network_region))


def network_from_network_code(network_code: str) -> NetworkSchema:
    network_code = network_code.upper().strip()

    if network_code in ["AU"]:
        return NetworkAU

    if network_code in ["WEM"]:
        return NetworkWEM

    if network_code in ["NEM"]:
        return NetworkNEM

    if network_code in ["APVI"]:
        return NetworkAPVI

    network_lookup = list(filter(lambda n: n.code == network_code, NETWORKS))

    if len(network_lookup):
        return network_lookup.pop()

    raise Exception("Unknown network {}".format(network_code))



from opennem.pipelines.wem.facilities import (
    WemStoreFacility,
    WemStoreLiveFacilities,
)
from opennem.spiders.wem import WemCurrentSpider


class WemLiveFacilities(WemCurrentSpider):
    name = "au.wem.live.facilities"

    pipelines_extra = set([WemStoreLiveFacilities])

    start_url = (
        "https://aemo.com.au/aemo/data/wa/infographic/facility-meta.csv"
    )


class WemFacilities(WemCurrentSpider):
    name = "au.wem.facilities"
    start_url = (
        "http://data.wa.aemo.com.au/datafiles/facilities/facilities.csv"
    )

    pipelines_extra = set([WemStoreFacility])



# pylint: disable=no-member
"""Add timezone indexes to facility_scada

Revision ID: d45e70a09231
Revises: 42cdf37a6e3b
Create Date: 2020-10-06 14:05:47.917286

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d45e70a09231"
down_revision = "42cdf37a6e3b"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("COMMIT")
    op.create_index(
        "idx_facility_scada_trading_interval_perth_year",
        "facility_scada",
        [
            sa.text(
                "date_trunc('year', trading_interval AT TIME ZONE 'Australia/Perth')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_perth_month",
        "facility_scada",
        [
            sa.text(
                "date_trunc('month', trading_interval AT TIME ZONE 'Australia/Perth')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_perth_day",
        "facility_scada",
        [
            sa.text(
                "date_trunc('day', trading_interval AT TIME ZONE 'Australia/Perth')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_perth_hour",
        "facility_scada",
        [
            sa.text(
                "date_trunc('hour', trading_interval AT TIME ZONE 'Australia/Perth')"
            )
        ],
        postgresql_concurrently=True,
    )

    # NEM timezone indexes
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_year",
        "facility_scada",
        [
            sa.text(
                "date_trunc('year', trading_interval AT TIME ZONE 'Australia/Sydney')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_month",
        "facility_scada",
        [
            sa.text(
                "date_trunc('month', trading_interval AT TIME ZONE 'Australia/Sydney')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_day",
        "facility_scada",
        [
            sa.text(
                "date_trunc('day', trading_interval AT TIME ZONE 'Australia/Sydney')"
            )
        ],
        postgresql_concurrently=True,
    )
    op.create_index(
        "idx_facility_scada_trading_interval_sydney_hour",
        "facility_scada",
        [
            sa.text(
                "date_trunc('hour', trading_interval AT TIME ZONE 'Australia/Sydney')"
            )
        ],
        postgresql_concurrently=True,
    )


def downgrade():
    op.drop_index("idx_facility_scada_trading_interval_perth_year")
    op.drop_index("idx_facility_scada_trading_interval_perth_month")
    op.drop_index("idx_facility_scada_trading_interval_perth_day")
    op.drop_index("idx_facility_scada_trading_interval_perth_hour")

    op.drop_index("idx_facility_scada_trading_interval_sydney_year")
    op.drop_index("idx_facility_scada_trading_interval_sydney_month")
    op.drop_index("idx_facility_scada_trading_interval_sydney_day")
    op.drop_index("idx_facility_scada_trading_interval_sydney_hour")



# pylint: disable=no-member
"""
Stats table

Revision ID: 889a1453be90
Revises: b30e47defd26
Create Date: 2021-01-05 16:18:59.833323

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "889a1453be90"
down_revision = "b30e47defd26"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "stats",
        sa.Column("created_by", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column(
            "stat_date", postgresql.TIMESTAMP(timezone=True), nullable=False
        ),
        sa.Column("country", sa.Text(), nullable=False),
        sa.Column("stat_type", sa.Text(), nullable=False),
        sa.Column("value", sa.Numeric(), nullable=True),
        sa.PrimaryKeyConstraint("stat_date", "country", "stat_type"),
    )
    op.create_index(
        op.f("ix_stats_stat_date"), "stats", ["stat_date"], unique=False
    )


def downgrade() -> None:
    op.drop_index(op.f("ix_stats_stat_date"), table_name="stats")
    op.drop_table("stats")



from io import BytesIO

import scrapy
from openpyxl import load_workbook

from opennem.pipelines.aemo.registration_exemption import (
    RegistrationExemptionGrouperPipeline,
    RegistrationExemptionStorePipeline,
)


class AEMORegistrationExemptionListSpider(scrapy.Spider):
    """
    Crawls the current AEMO registration and exemption spreadsheet
    and extracts participants and generators

    """

    # name = "au.aemo.current.registration_exemption"

    start_urls = [
        "https://data.opennem.org.au/v3/data/NEM+Registration+and+Exemption+List+July.xlsx"
    ]

    participant_keys = ["name", "abn"]

    generator_keys = [
        "participant",
        "station_name",
        "region",
        "dispatch_type",
        "category",
        "classification",
        "fuel_source_primary",
        "fuel_source_descriptor",
        "tech_primary",
        "tech_primary_descriptor",
        "unit_no",
        "unit_size",
        "aggreagation",
        "duid",
        "reg_cap",
        "max_cap",
        "max_roc",
    ]

    pipelines_extra = set(
        [
            RegistrationExemptionGrouperPipeline,
            RegistrationExemptionStorePipeline,
        ]
    )

    def parse(self, response):

        wb = load_workbook(BytesIO(response.body), data_only=True)

        generator_ws = wb.get_sheet_by_name("Generators and Scheduled Loads")
        participant_ws = wb.get_sheet_by_name("Registered Participants")

        generators = []
        participants = []

        for row in generator_ws.iter_rows(min_row=2, values_only=True):
            generators.append(
                dict(
                    zip(
                        self.generator_keys,
                        list(row[0 : len(self.generator_keys)]),
                    )
                )
            )

        for row in participant_ws.iter_rows(min_row=3, values_only=True):

            participants.append(
                dict(
                    zip(
                        self.participant_keys,
                        list(row[0 : len(self.participant_keys)]),
                    )
                )
            )

        participants = sorted(participants, key=lambda k: k["name"])

        yield {
            "generators": generators,
            "participants": participants,
        }



# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument
# pylint: disable=no-member
import platform

from huey import PriorityRedisHuey, crontab

from opennem.db.tasks import refresh_material_views, refresh_timescale_views
from opennem.settings import settings

# Py 3.8 on MacOS changed the default multiprocessing model
if platform.system() == "Darwin":
    import multiprocessing

    try:
        multiprocessing.set_start_method("fork")
    except RuntimeError:
        # sometimes it has already been set by
        # other libs
        pass

redis_host = None

if settings.cache_url:
    redis_host = settings.cache_url.host

huey = PriorityRedisHuey("opennem.scheduler.db", host=redis_host)


# database tasks
@huey.periodic_task(crontab(hour="*/1"))
@huey.lock_task("db_refresh_ts_views")
def db_refresh_ts_views() -> None:
    refresh_timescale_views()


@huey.periodic_task(crontab(hour="*/6"))
@huey.lock_task("db_refresh_material_views")
def db_refresh_material_views() -> None:
    refresh_material_views()


@huey.periodic_task(crontab(hour="*/6"), priority=10)
@huey.lock_task("db_refresh_all")
def db_refresh_all() -> None:
    refresh_timescale_views(True)



# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument
# pylint: disable=no-member
import platform

from huey import PriorityRedisHuey, crontab

from opennem.api.export.map import PriorityType
from opennem.api.export.tasks import (
    export_all_daily,
    export_all_monthly,
    export_energy,
    export_metadata,
    export_power,
)
from opennem.exporter.geojson import export_facility_geojson
from opennem.monitors.aemo_intervals import aemo_wem_live_interval
from opennem.monitors.opennem import check_opennem_interval_delays
from opennem.monitors.opennem_metadata import check_metadata_status
from opennem.notifications.slack import slack_message
from opennem.settings import settings

# Py 3.8 on MacOS changed the default multiprocessing model
if platform.system() == "Darwin":
    import multiprocessing

    try:
        multiprocessing.set_start_method("fork")
    except RuntimeError:
        # sometimes it has already been set by
        # other libs
        pass

redis_host = None

if settings.cache_url:
    redis_host = settings.cache_url.host

huey = PriorityRedisHuey("opennem.scheduler", host=redis_host)
# export tasks


@huey.periodic_task(crontab(minute="*/3"), priority=90)
@huey.lock_task("schedule_live_tasks")
def schedule_live_tasks() -> None:
    export_power(priority=PriorityType.live)


# @huey.periodic_task(crontab(hour="*/12"))
# @huey.lock_task("schedule_power_weeklies")
# def schedule_power_weeklies() -> None:
#     """
#     Run weekly power outputs
#     """
#     export_power(priority=PriorityType.history, latest=True)


# @huey.periodic_task(crontab(minute="15", hour="12"))
# @huey.lock_task("schedule_power_weeklies_archive")
# def schedule_power_weeklies_archive() -> None:
#     """
#     Run weekly power outputs entire archive
#     """
#     export_power(priority=PriorityType.history)


@huey.periodic_task(crontab(minute="*/30"), priority=50)
@huey.lock_task("schedule_export_all_daily")
def schedule_export_all_daily() -> None:
    export_all_daily()
    slack_message("Finished running export_all_daily")


@huey.periodic_task(crontab(minute="*/30"), priority=50)
@huey.lock_task("schedule_export_all_monthly")
def schedule_export_all_monthly() -> None:
    export_all_monthly()


@huey.periodic_task(crontab(minute="*/5"), priority=50)
@huey.lock_task("schedule_hourly_tasks")
def schedule_hourly_tasks() -> None:
    export_energy(priority=PriorityType.daily, latest=True)


@huey.periodic_task(crontab(minute="*/15"))
@huey.lock_task("schedule_daily_tasks")
def schedule_daily_tasks() -> None:
    export_energy(priority=PriorityType.daily)
    slack_message("Finished running energy dailies")


@huey.periodic_task(crontab(minute="*/15"), priority=30)
@huey.lock_task("schedule_energy_monthlies")
def schedule_energy_monthlies() -> None:
    export_energy(priority=PriorityType.monthly)
    slack_message("Finished running energy_monthlies")


# geojson maps
@huey.periodic_task(crontab(minute="*/5"), priority=50)
@huey.lock_task("schedule_export_geojson")
def schedule_export_geojson() -> None:
    export_facility_geojson()


# metadata
@huey.periodic_task(crontab(minute="*/30"), priority=30)
@huey.lock_task("schedule_export_metadata")
def schedule_export_metadata() -> None:
    export_metadata()


# monitoring tasks
@huey.periodic_task(crontab(minute="*/15"), priority=80)
@huey.lock_task("monitor_opennem_intervals")
def monitor_opennem_intervals() -> None:
    for network_code in ["NEM", "WEM"]:
        check_opennem_interval_delays(network_code)


@huey.periodic_task(crontab(minute="*/15"), priority=50)
@huey.lock_task("monitor_wem_interval")
def monitor_wem_interval() -> None:
    aemo_wem_live_interval()


@huey.periodic_task(crontab(hour="*/12"), priority=10)
@huey.lock_task("monitor_metadata_status")
def monitor_metadata_status() -> None:
    check_metadata_status()



from typing import List, Optional, Union

from pydantic import BaseModel, Field


class ApiBase(BaseModel):
    class Config:
        orm_mode = True
        anystr_strip_whitespace = True
        use_enum_values = True
        arbitrary_types_allowed = True
        validate_assignment = True


class UpdateResponse(BaseModel):
    success: bool = True
    records: List = []


class FueltechResponse(ApiBase):
    success: bool = True

    # @TODO fix circular references
    # records: List[FueltechSchema]




#!/usr/bin/env python
"""
Srapyd control methods

"""
import logging
from typing import Any, Dict, List
from urllib.parse import urljoin

import requests

from opennem.settings import settings
from opennem.utils.scrapy import get_spiders

logger = logging.getLogger("scrapyd.client")


def get_jobs() -> Dict[str, Any]:
    job_url = urljoin(
        settings.scrapyd_url,
        "listjobs.json?project={}".format(settings.scrapyd_project_name),
    )

    jobs = requests.get(job_url).json()

    return jobs


def job_cancel(id: str) -> bool:
    cancel_job_url = urljoin(settings.scrapyd_url, "cancel.json")

    r = requests.post(cancel_job_url, data={"project": "opennem", "job": id})

    resp = r.json()

    if resp["status"] == "error":
        logger.error("Error: {}".format(resp["message"]))
        return False

    logger.info("Cancelled job: {}".format(resp["jobid"]))

    return True


def job_schedule(spider_name: str) -> bool:
    schedule_url = urljoin(settings.scrapyd_url, "schedule.json")

    r = requests.post(
        schedule_url, data={"project": "opennem", "spider": spider_name}
    )

    if not r.ok:
        logger.error("Error: {}".format(r.status_code))
        return False

    resp = r.json()

    if resp["status"] == "error":
        logger.error("Error: {}".format(resp["message"]))
        return False

    logger.info(
        "Queued spider {} with task: {}".format(spider_name, resp["jobid"])
    )

    return True


def job_cancel_state(state: str = "pending") -> bool:
    jobs = get_jobs()

    if state not in jobs:
        logger.info("Invalid state or no jobs in state {}".format(state))
        return False

    pending_jobs = jobs[state]

    for job in pending_jobs:
        job_id = job["id"]
        logger.info("Cancelling {}".format(job_id))
        job_cancel(job_id)

    return True


def job_schedule_all(matches: str = None) -> List[str]:
    spiders = get_spiders()

    spider_scheduled = []

    for s in spiders:
        if matches and matches not in s:
            continue
        job_schedule(s)
        spider_scheduled.append(s)

    return spider_scheduled



"""
Methods to store and persist stats in the database or to JSON
"""
import logging

from sqlalchemy.dialects.postgresql import insert

from opennem.core.loader import get_data_path
from opennem.core.parsers.cpi import stat_au_cpi
from opennem.db import SessionLocal, get_database_engine
from opennem.db.models.opennem import Stats
from opennem.schema.stats import StatsSet

logger = logging.getLogger("opennem.stats.store")


def store_stats_database(statset: StatsSet) -> int:
    s = SessionLocal()

    records_to_store = [i.dict() for i in statset.stats]

    stmt = insert(Stats).values(records_to_store)
    stmt.bind = get_database_engine()
    stmt = stmt.on_conflict_do_update(
        index_elements=[
            "stat_date",
            "country",
            "stat_type",
        ],
        set_={
            "value": stmt.excluded.value,
        },
    )

    try:
        s.execute(stmt)
        s.commit()
    except Exception as e:
        logger.error("Error inserting records")
        logger.error(e)
        return 0
    finally:
        s.close()

    num_records = len(records_to_store)

    logger.info("Wrote {} records to database".format(num_records))

    return num_records


def store_stats_json(records_to_store: StatsSet) -> bool:
    data_dir = get_data_path()

    with open(data_dir / "au_cpi.json", "w") as fh:
        fh.write(records_to_store.json(indent=4))

    logger.info("Wrote stat set json to {}".format(data_dir / "au_cpi.json"))

    return True


if __name__ == "__main__":
    r = stat_au_cpi()
    store_stats_database(r)



import logging
from typing import Generator

from databases import Database
from sqlalchemy import create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from opennem.exporter.encoders import opennem_deserialize, opennem_serialize
from opennem.settings import settings

DeclarativeBase = declarative_base()

logger = logging.getLogger(__name__)


database = Database(settings.db_url)


# Methods


async def db_connect_async() -> None:
    await database.connect()


async def db_disconnect() -> None:
    await database.disconnect()


def db_connect(
    db_name: str = None, debug: bool = False, timeout: int = 300
) -> Engine:
    """
    Performs database connection using database settings from settings.py.

    Returns sqlalchemy engine instance
    """
    db_conn_str = settings.db_url

    connect_args = {}

    if db_conn_str.startswith("sqlite"):
        connect_args = {"check_same_thread": False}

    if settings.db_debug:
        debug = True

    try:
        return create_engine(
            db_conn_str,
            json_serializer=opennem_serialize,
            json_deserializer=opennem_deserialize,
            echo=debug,
            pool_size=10,
            max_overflow=2,
            pool_recycle=300,
            pool_timeout=timeout,
            pool_pre_ping=True,
            pool_use_lifo=True,
            connect_args=connect_args,
        )
    except Exception as exc:
        logger.error("Could not connect to database: %s", exc)
        raise exc


engine = db_connect()

SessionLocal = sessionmaker(
    bind=engine,
    autocommit=False,
    autoflush=False,
)

SessionAutocommit = sessionmaker(bind=engine, autocommit=True, autoflush=True)


def get_database_session() -> Generator[sessionmaker, None, None]:
    """
    Gets a database session

    """
    s = None

    try:
        s = SessionLocal()
        yield s
    except Exception as e:
        raise e
    finally:
        if s:
            s.close()


def get_database_engine() -> Engine:
    """
    Gets a database engine connection

    """
    return engine



"""
    Generates the map of queries and paths to export to S3 buckets
    for opennem.org.au

"""

import logging
from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel

from opennem.api.stats.controllers import ScadaDateRange, get_scada_range
from opennem.api.time import human_to_period
from opennem.core.network_region_bom_station_map import (
    get_network_region_weather_station,
)
from opennem.core.networks import (
    NetworkAPVI,
    NetworkAU,
    NetworkNEM,
    NetworkSchema,
    NetworkWEM,
    network_from_network_code,
)
from opennem.db import SessionLocal
from opennem.db.models.opennem import Network
from opennem.schema.time import TimePeriod
from opennem.utils.dates import date_range_from_week, week_series
from opennem.utils.version import VersionPart, get_version

logger = logging.getLogger(__name__)

VERSION_MAJOR = get_version(version_part=VersionPart.MAJOR)
STATS_FOLDER = "stats"


class StatType(Enum):
    power = "power"
    energy = "energy"
    gov = "gov"


class PriorityType(Enum):
    live = 1
    daily = 2
    monthly = 3
    history = 4


def priority_from_name(priority_name: str) -> PriorityType:
    # @TODO this doesn't feel right
    if priority_name == "live":
        return PriorityType.live
    if priority_name == "daily":
        return PriorityType.daily
    if priority_name == "monthly":
        return PriorityType.monthly
    if priority_name == "history":
        return PriorityType.history

    raise Exception("Could not find priority: {}".format(priority_name))


class StatExport(BaseModel):
    stat_type: StatType = StatType.energy
    priority: PriorityType = PriorityType.live
    country: str
    network: NetworkSchema
    networks: Optional[List[NetworkSchema]]
    network_region: Optional[str]
    network_region_query: Optional[str]
    date_range: ScadaDateRange
    bom_station: Optional[str]
    year: Optional[int]
    week: Optional[int]
    period: Optional[TimePeriod]
    file_path: Optional[str]

    @property
    def path(self) -> str:
        _path_components = [
            f"v{VERSION_MAJOR}",
            STATS_FOLDER,
            self.country,
            self.network.code,
        ]

        if self.network_region:
            _path_components.append(self.network_region)

        _path_components.append(self.stat_type.value)

        if self.period:
            _path_components.append(self.period.period_human)

        if self.week:
            _path_components.append("week")

        if self.year:
            _path_components.append(str(self.year))

        if self.week:
            _path_components.append(str(self.week))

        dir_path = "/".join([str(i) for i in _path_components])

        return "{}.json".format(dir_path)


class StatMetadata(BaseModel):
    date_created: datetime
    version: Optional[str]
    resources: List[StatExport]

    def get_by_stat_type(
        self, stat_type: StatType, priority: Optional[PriorityType] = None
    ) -> List[StatExport]:
        if priority:
            return list(
                filter(
                    lambda s: s.stat_type == stat_type
                    and s.priority == priority,
                    self.resources,
                )
            )

        return list(filter(lambda s: s.stat_type == stat_type, self.resources))


def get_export_map() -> StatMetadata:
    """
    Generates a map of all export JSONs

    """
    session = SessionLocal()

    networks = (
        session.query(Network).filter(Network.export_set.is_(True)).all()
    )

    if not networks:
        raise Exception("No networks")

    countries = list(set([network.country for network in networks]))

    _exmap = []

    for country in countries:
        # @TODO derive this
        scada_range = get_scada_range(
            network=NetworkAU, networks=[NetworkNEM, NetworkWEM]
        )

        if not scada_range:
            raise Exception("Require a scada range")

        export = StatExport(
            stat_type=StatType.power,
            priority=PriorityType.live,
            country=country,
            date_range=scada_range,
            network=NetworkAU,
            networks=[NetworkNEM, NetworkWEM],
            period=human_to_period("7d"),
        )

        _exmap.append(export)

        for year, week in week_series(scada_range.end, scada_range.start):
            export = StatExport(
                stat_type=StatType.power,
                priority=PriorityType.history,
                country=country,
                network=NetworkAU,
                networks=[NetworkNEM, NetworkWEM],
                year=year,
                week=week,
                date_range=date_range_from_week(year, week, NetworkAU),
            )
            _exmap.append(export)

        for year in range(
            datetime.now().year,
            scada_range.start.year - 1,
            -1,
        ):
            export = StatExport(
                stat_type=StatType.energy,
                priority=PriorityType.daily,
                country=country,
                date_range=scada_range,
                network=NetworkAU,
                networks=[NetworkNEM, NetworkWEM],
                year=year,
            )
            _exmap.append(export)

        export = StatExport(
            stat_type=StatType.energy,
            priority=PriorityType.monthly,
            country=country,
            date_range=scada_range,
            network=NetworkAU,
            networks=[NetworkNEM, NetworkWEM],
            period=human_to_period("all"),
        )
        _exmap.append(export)

    for network in networks:
        network_schema = network_from_network_code(network.code)
        scada_range = get_scada_range(network=network_schema)
        bom_station = get_network_region_weather_station(network.code)

        export = StatExport(
            stat_type=StatType.power,
            priority=PriorityType.live,
            country=network.country,
            date_range=scada_range,
            network=network_schema,
            bom_station=bom_station,
            period=human_to_period("7d"),
        )

        if network.code == "WEM":
            export.networks = [NetworkWEM, NetworkAPVI]
            export.network_region_query = "WEM"

        _exmap.append(export)

        if not scada_range:
            raise Exception("Require a scada range")

        for year, week in week_series(scada_range.end, scada_range.start):
            export = StatExport(
                stat_type=StatType.power,
                priority=PriorityType.history,
                country=network.country,
                network=network_schema,
                year=year,
                week=week,
                date_range=date_range_from_week(year, week, NetworkAU),
            )

            if network.code == "WEM":
                export.networks = [NetworkWEM, NetworkAPVI]
                export.network_region_query = "WEM"

            _exmap.append(export)

        for year in range(
            datetime.now().year,
            scada_range.start.year - 1,
            -1,
        ):
            export = StatExport(
                stat_type=StatType.energy,
                priority=PriorityType.daily,
                country=network.country,
                date_range=scada_range,
                network=network_schema,
                bom_station=bom_station,
                year=year,
            )

            if network.code == "WEM":
                export.networks = [NetworkWEM, NetworkAPVI]
                export.network_region_query = "WEM"

            _exmap.append(export)

        export = StatExport(
            stat_type=StatType.energy,
            priority=PriorityType.monthly,
            country=network.country,
            date_range=scada_range,
            network=network_schema,
            bom_station=bom_station,
            period=human_to_period("all"),
        )

        if network.code == "WEM":
            export.networks = [NetworkWEM, NetworkAPVI]
            export.network_region_query = "WEM"

        _exmap.append(export)

        # Skip cases like wem/wem where region is supurfelous
        if len(network.regions) < 2:
            continue

        for region in network.regions:
            scada_range = get_scada_range(
                network=network_schema, network_region=region
            )
            bom_station = get_network_region_weather_station(region.code)

            if not scada_range:
                raise Exception("Require a scada range")

            export = StatExport(
                stat_type=StatType.power,
                priority=PriorityType.live,
                country=network.country,
                date_range=scada_range,
                network=network_schema,
                network_region=region.code,
                bom_station=bom_station,
                period=human_to_period("7d"),
            )

            if network.code == "WEM":
                export.networks = [NetworkWEM, NetworkAPVI]
                export.network_region_query = "WEM"

            _exmap.append(export)

            for year in range(
                datetime.now().year,
                scada_range.start.year - 1,
                -1,
            ):
                export = StatExport(
                    stat_type=StatType.energy,
                    priority=PriorityType.daily,
                    country=network.country,
                    date_range=scada_range,
                    network=network_schema,
                    network_region=region.code,
                    bom_station=bom_station,
                    year=year,
                )
                _exmap.append(export)

            export = StatExport(
                stat_type=StatType.energy,
                priority=PriorityType.monthly,
                country=network.country,
                date_range=scada_range,
                network=network_schema,
                network_region=region.code,
                bom_station=bom_station,
                period=human_to_period("all"),
            )

            if network.code == "WEM":
                export.networks = [NetworkWEM, NetworkAPVI]
                export.network_region_query = "WEM"

            _exmap.append(export)

    export_meta = StatMetadata(
        date_created=datetime.now(), version=get_version(), resources=_exmap
    )

    return export_meta


if __name__ == "__main__":
    pass



# pylint: disable=no-member
"""
Facility scada network index

Revision ID: bfb98983b92c
Revises: ee93cbe26f74
Create Date: 2020-12-03 06:05:10.081580

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "bfb98983b92c"
down_revision = "ee93cbe26f74"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "create index if not exists idx_facility_scada_network_id on facility_scada  (network_id)"
    )


def downgrade() -> None:
    op.execute("drop index idx_facility_scada_network_id")



"""
Tasks file to expor JSON's to S3 or locally for the
opennem website

This is the most frequently accessed content that doesn't
require the API


"""

import logging
from datetime import datetime
from typing import List, Optional

from opennem.api.export.controllers import (
    energy_fueltech_daily,
    energy_interconnector_emissions_region_daily,
    energy_interconnector_region_daily,
    gov_stats_cpi,
    power_flows_week,
    power_week,
    weather_daily,
)
from opennem.api.export.map import PriorityType, StatExport, StatType, get_export_map
from opennem.api.export.utils import write_output
from opennem.api.stats.controllers import get_scada_range
from opennem.api.stats.schema import OpennemDataSet
from opennem.core.network_region_bom_station_map import get_network_region_weather_station
from opennem.core.networks import network_from_network_code
from opennem.db import SessionLocal
from opennem.db.models.opennem import NetworkRegion
from opennem.schema.network import NetworkAPVI, NetworkNEM, NetworkWEM
from opennem.settings import settings
from opennem.utils.version import get_version

logger = logging.getLogger(__name__)


def export_power(
    stats: List[StatExport] = None,
    priority: Optional[PriorityType] = None,
    latest: Optional[bool] = False,
) -> None:
    """
    Export power stats from the export map


    """
    if not stats:
        export_map = get_export_map()
        stats = export_map.get_by_stat_type(StatType.power, priority)

    output_count: int = 0

    for power_stat in stats:
        if power_stat.stat_type != StatType.power:
            continue

        if output_count >= 1 and latest:
            return None

        stat_set = power_week(
            date_range=power_stat.date_range,
            period=power_stat.period,
            network_code=power_stat.network.code,
            network_region_code=power_stat.network_region_query or power_stat.network_region,
            networks_query=power_stat.networks,
        )

        if not stat_set:
            logger.info(
                "No power stat set for {} {} {}".format(
                    power_stat.period,
                    power_stat.networks,
                    power_stat.network_region,
                )
            )
            continue

        if power_stat.network_region:
            flow_set = power_flows_week(
                network=NetworkNEM,
                date_range=power_stat.date_range,
                network_region_code=power_stat.network_region,
            )

            if flow_set:
                stat_set.append_set(flow_set)

        if power_stat.bom_station:
            weather_set = weather_daily(
                station_code=power_stat.bom_station,
                network_code=power_stat.network.code,
                include_min_max=False,
                period_human="7d",
                unit_name="temperature",
            )

            stat_set.append_set(weather_set)

        write_output(power_stat.path, stat_set)
        output_count += 1


def export_energy(
    stats: List[StatExport] = None,
    priority: Optional[PriorityType] = None,
    latest: Optional[bool] = False,
) -> None:
    """
    Export energy stats from the export map


    """
    if not stats:
        export_map = get_export_map()
        stats = export_map.get_by_stat_type(StatType.energy, priority)

    CURRENT_YEAR = datetime.now().year

    for energy_stat in stats:
        if energy_stat.stat_type != StatType.energy:
            continue

        # @FIX trim to NEM since it's the one with the shortest
        # data time span.
        # @TODO find a better and more flexible way to do this in the
        # range method
        date_range_networks = energy_stat.networks or []

        if NetworkNEM in date_range_networks:
            date_range_networks = [NetworkNEM]

        date_range = get_scada_range(network=energy_stat.network, networks=date_range_networks)

        if energy_stat.year:

            if latest and energy_stat.year != CURRENT_YEAR:
                continue

            stat_set = energy_fueltech_daily(
                interval_size="1d",
                year=energy_stat.year,
                network=energy_stat.network,
                networks_query=energy_stat.networks,
                date_range=date_range,
                network_region_code=energy_stat.network_region_query or energy_stat.network_region,
            )

            if not stat_set:
                continue

            # Hard coded to NEM only atm but we'll put has_interconnectors
            # in the metadata to automate all this
            if energy_stat.network == NetworkNEM and energy_stat.network_region:
                interconnector_flows = energy_interconnector_region_daily(
                    interval_size="1d",
                    year=energy_stat.year,
                    network=energy_stat.network,
                    networks_query=energy_stat.networks,
                    date_range=date_range,
                    network_region_code=energy_stat.network_region_query
                    or energy_stat.network_region,
                )
                stat_set.append_set(interconnector_flows)

                interconnector_emissions = energy_interconnector_emissions_region_daily(
                    interval_size="1d",
                    year=energy_stat.year,
                    network=energy_stat.network,
                    networks_query=energy_stat.networks,
                    date_range=date_range,
                    network_region_code=energy_stat.network_region_query
                    or energy_stat.network_region,
                )
                stat_set.append_set(interconnector_emissions)

            if energy_stat.bom_station:
                weather_stats = weather_daily(
                    station_code=energy_stat.bom_station,
                    year=energy_stat.year,
                    network_code=energy_stat.network.code,
                    network_region=energy_stat.network_region,
                )
                stat_set.append_set(weather_stats)

            write_output(energy_stat.path, stat_set)

        elif energy_stat.period and energy_stat.period.period_human == "all":

            stat_set = energy_fueltech_daily(
                interval_size="1M",
                network=energy_stat.network,
                networks_query=energy_stat.networks,
                date_range=date_range,
                network_region_code=energy_stat.network_region_query or energy_stat.network_region,
            )

            if not stat_set:
                continue

            # Hard coded to NEM only atm but we'll put has_interconnectors
            # in the metadata to automate all this
            if energy_stat.network == NetworkNEM and energy_stat.network_region:
                interconnector_flows = energy_interconnector_region_daily(
                    interval_size="1M",
                    network=energy_stat.network,
                    networks_query=energy_stat.networks,
                    date_range=date_range,
                    network_region_code=energy_stat.network_region_query
                    or energy_stat.network_region,
                )
                stat_set.append_set(interconnector_flows)

                interconnector_emissions = energy_interconnector_emissions_region_daily(
                    interval_size="1M",
                    network=energy_stat.network,
                    networks_query=energy_stat.networks,
                    date_range=date_range,
                    network_region_code=energy_stat.network_region_query
                    or energy_stat.network_region,
                )
                stat_set.append_set(interconnector_emissions)

            if energy_stat.bom_station:
                weather_stats = weather_daily(
                    station_code=energy_stat.bom_station,
                    date_range=date_range,
                    year=energy_stat.year,
                    network_code=energy_stat.network.code,
                    network_region=energy_stat.network_region,
                )
                stat_set.append_set(weather_stats)

            write_output(energy_stat.path, stat_set)


def export_all_monthly() -> None:
    session = SessionLocal()
    network_regions = session.query(NetworkRegion).all()

    all_monthly = OpennemDataSet(
        code="au", data=[], version=get_version(), created_at=datetime.now()
    )

    cpi = gov_stats_cpi()
    all_monthly.append_set(cpi)

    for network_region in network_regions:
        network = network_from_network_code(network_region.network.code)
        networks = None

        if network_region.code == "WEM":
            networks = [NetworkWEM, NetworkAPVI]

        stat_set = energy_fueltech_daily(
            interval_size="1M",
            network=network,
            networks_query=networks,
            network_region_code=network_region.code,
        )

        if not stat_set:
            continue

        if network == NetworkNEM:
            interconnector_flows = energy_interconnector_region_daily(
                interval_size="1M",
                network=network,
                networks_query=networks,
                network_region_code=network_region.code,
            )
            stat_set.append_set(interconnector_flows)

            interconnector_emissions = energy_interconnector_emissions_region_daily(
                interval_size="1M",
                network=network,
                networks_query=networks,
                network_region_code=network_region.code,
            )
            stat_set.append_set(interconnector_emissions)

        all_monthly.append_set(stat_set)

        bom_station = get_network_region_weather_station(network_region.code)

        if bom_station:
            weather_stats = weather_daily(
                station_code=bom_station, network_code=network.code, network_region=network_region
            )
            all_monthly.append_set(weather_stats)

    write_output("v3/stats/au/all/monthly.json", all_monthly)


def export_all_daily() -> None:
    session = SessionLocal()
    network_regions = session.query(NetworkRegion).all()

    cpi = gov_stats_cpi()

    for network_region in network_regions:
        network = network_from_network_code(network_region.network.code)
        networks = None

        if network_region.code == "WEM":
            networks = [NetworkWEM, NetworkAPVI]

        stat_set = energy_fueltech_daily(
            interval_size="1d",
            network=network,
            networks_query=networks,
            network_region_code=network_region.code,
        )

        if not stat_set:
            continue

        # Hard coded to NEM only atm but we'll put has_interconnectors
        # in the metadata to automate all this
        if network == NetworkNEM:
            interconnector_flows = energy_interconnector_region_daily(
                interval_size="1d",
                network=network,
                networks_query=networks,
                network_region_code=network_region.code,
            )
            stat_set.append_set(interconnector_flows)

            interconnector_emissions = energy_interconnector_emissions_region_daily(
                interval_size="1d",
                network=network,
                networks_query=networks,
                network_region_code=network_region.code,
            )
            stat_set.append_set(interconnector_emissions)

        bom_station = get_network_region_weather_station(network_region.code)

        if bom_station:
            weather_stats = weather_daily(
                station_code=bom_station, network_code=network.code, network_region=network_region
            )
            stat_set.append_set(weather_stats)

        if cpi:
            stat_set.append_set(cpi)

        write_output(f"v3/stats/au/{network_region.code}/daily.json", stat_set)


def export_metadata() -> bool:
    """
    Export metadata


    """
    _export_map_out = get_export_map()

    # this is a hack because pydantic doesn't
    # serialize properties
    for r in _export_map_out.resources:
        r.file_path = r.path

    wrote_bytes = write_output("metadata.json", _export_map_out)

    if wrote_bytes and wrote_bytes > 0:
        return True

    return False


if __name__ == "__main__":
    export_all_daily()
    # export_all_monthly()
    # export_energy()
    # export_all_daily()

    # export_power(priority=PriorityType.history)

    # if settings.env in ["development", "staging"]:
    #     export_power(priority=PriorityType.live)
    #     export_energy(latest=True)
    #     export_metadata()
    #     # export_power(priority=PriorityType.history)
    # else:
    #     # export_power(priority=PriorityType.live)
    #     # export_energy(latest=True)
    #     export_energy()
    #     # export_metadata()
    #     # export_power(priority=PriorityType.history)



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel, Field

from opennem.api.facility.schema import FacilityRecord
from opennem.api.schema import ApiBase
from opennem.api.station.schema import StationRecord

# from opennem.schema.core import BaseConfig
from opennem.schema.opennem import LocationSchema, RevisionSchema


class UpdateResponse(BaseModel):
    success: bool = True
    records: List = []


class RevisionModificationTypes(str, Enum):
    approve = "approve"
    reject = "reject"


class RevisionModification(ApiBase):
    comment: Optional[str] = Field(None)
    modification: RevisionModificationTypes


class RevisionModificationResponse(ApiBase):
    success: bool = False
    record: RevisionSchema
    target: Optional[Union[FacilityRecord, StationRecord, LocationSchema]]




from opennem.core.loader import load_data

from .mms import run_import_mms


def run_import_opennem_registry():
    station_fixture = load_data("facility_registry.json", from_fixture=True)
    mms = run_import_mms()

    nem_stations = {
        i: v
        for i, v in station_fixture.items()
        if v["location"]["state"] not in ["WA"]
    }

    for station_key, station_record in nem_stations.items():
        if not station_key in mms:
            print("Got {} which is not in registry".format(station_key))


if __name__ == "__main__":
    run_import_opennem_registry()



import json
import os

TEST_FIXTURE_PATH = os.path.realpath(
    os.path.join(os.path.dirname(__file__), "..", "..", "data")
)


def load_fixture(fixture_name):
    fixture_path = os.path.join(TEST_FIXTURE_PATH, fixture_name)

    if not os.path.isfile(fixture_path):
        raise Exception("Not a file: {}".format(fixture_path))

    fixture = None

    with open(fixture_path) as fh:
        fixture = json.load(fh)

    return fixture
